/* *** CfgPool.c *** Configuration Pool

              Copyright (C) 2002 Ral Nez de Arenas Coronado

        This program is free software; you can redistribute it and/or
         modify it under the terms of the GNU General Public License
                as published by the Free Software Foundation;
                      either version 2 of the License,
                   or (at your option) any later version.

       This program is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty
           of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
            See the GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
             (`GPL') along with this program; if not, write to:

                       Free Software Foundation, Inc.
                         59 Temple Place, Suite 330
                         Boston, MA 02111-1307  USA

                     Report bugs to <raul@pleyades.net>
*/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#include <sys/stat.h>
#include <inttypes.h>
#include "cfgpool.h"
#include "mobs.h"


/* NOTE: This is NOT a class... */
struct CfgItem {
    char *key;
    char *val;
    struct CfgItem *nxt;
};

/* This is the class... */
struct CfgPool {
    struct CfgItem *stack;
};




/*
    This function reads the contents of 'filename' into core, parses it
and stores the configuration items found into the 'pool'.

    The configuration file lines can be empty lines (those containing just
whitespaces), comment lines (those containing whitespaces followed by a '#'
and some text) and key-value lines. The key value lines consist of a key
(that is, some text with no whitespaces interspersed), some whitespace
following it and the value (some arbitrary text). The key-value pair can
have whitespace before and after: it is removed when reading.

    This function returns NULL on error, and a CfgPool object otherwise.
    
    ENOENT: Cannot find or open the file.
    E2BIG: The file is too big.
    ENOMEM: Not enough memory to build object.
    EIO: Problems reading the file.
    EINVAL: Ill formed line.
*/
CfgPool cfgpool_create (const char *filename) {

    char *buffer=NULL;  /* Buffer containing file pages */
    size_t buffsize=0;
    struct stat st;
    int fd=-1;
    CfgPool pool;  /* The pool itself */
    
    ASSERT(filename);
    
    if (stat(filename, &st) < 0) {
        errno=ENOENT;  /* Doesn't exist :(( */
        return NULL;
    }
    if (st.st_size > SSIZE_MAX) {
        errno=E2BIG;  /* Such a big file!!! */
        return NULL;
    }

    buffsize=st.st_size;

    if ((fd=open(filename, O_RDONLY)) < 0) {
        errno=ENOENT;  /* Fuck, cannot open the file */
        return NULL;
    };

    buffer=malloc(buffsize);
    if (!buffer) {
        close(fd);
        errno=ENOMEM;
        return NULL;
    }
    if (read(fd, buffer, buffsize) < 0) {
        close(fd);
        errno=EIO;  /* Cannot read, what a pity... */
        return NULL;
    }
    close(fd);

    /* Well, build the object... */
    pool=malloc(sizeof(struct CfgPool));
    if (!pool) {
        free(buffer);
        return NULL;
    }

    pool->stack=NULL;

    /* The file is correctly mapped in 'buffer' */
    while (1) {  /* Why not... */
        static size_t current=0;
        struct CfgItem *item=NULL;
        size_t tmpindex=0;  /* Temporary index to parse lines */
        size_t bol=0;   /* Begin of line */
        size_t eol=0;   /* End of line */

        if (current >= buffsize) break;  /* No more lines... */

        bol=eol=current;  /* Initialize indexes */

        /* Find the end of the line */
        while ((eol < buffsize) && (buffer[eol] != '\n')) eol++;
        
        /* Well, we got a line: now process it*/
        tmpindex=bol;
        current=eol;
        current++;
        
        /* Convert '\0' to spaces in the line */
        while (tmpindex < eol) {
            if (!buffer[tmpindex]) buffer[tmpindex]=' ';
            tmpindex++;
        }
        tmpindex=bol;
        
        /* Skip leading and trailing spaces, if any */
        while ((tmpindex < eol) && isspace(buffer[tmpindex])) tmpindex++;
        if (tmpindex==eol) continue;  /* Empty line, get next */
        bol=tmpindex; tmpindex=eol;
        while ((tmpindex > bol) && isspace(buffer[tmpindex])) tmpindex--;
        tmpindex++; eol=tmpindex;

        if (buffer[bol]=='#') continue;  /* Comment line, skip */

        /* Now get the key */
        tmpindex=bol;
        while ((tmpindex < eol) && !isspace(buffer[tmpindex])) tmpindex++;
        if (tmpindex==eol) continue;  /* No value for this key, get next */

        /* Create new item to store key */
        item=malloc(sizeof(struct CfgItem));
        if (!item) {
            free(buffer);
            cfgpool_delete(pool);
            errno=ENOMEM;
            return NULL;
        }

        item->key=NULL;
        item->val=NULL;
        item->nxt=NULL;

        /* Allocate space for the key */
        item->key=malloc(tmpindex-bol+1);
        if (!(item->key)) {
            free(buffer);
            cfgpool_delete(pool);
            errno=ENOMEM;
            return NULL;
        }

        /* Copy the key */
        memcpy(item->key, &buffer[bol], tmpindex-bol);
        item->key[tmpindex-bol]='\0';

        
        /* Now get the value */
        bol=++tmpindex;

        /* Strip whitespaces after key */
        while ((tmpindex < eol) && isspace(buffer[tmpindex])) tmpindex++;
        bol=tmpindex;

        /* Allocate space for the val */
        item->val=malloc(eol-bol+1);
        if (!(item->val)) {
            free(buffer);
            cfgpool_delete(pool);
            errno=ENOMEM;
            return NULL;
        }

        /* Copy the val */
        memcpy(item->val, &buffer[bol], eol-bol);
        item->val[eol-bol]='\0';
        
        /* Push item onto the stack*/
        item->nxt=pool->stack;
        pool->stack=item;
    }
    
    free(buffer);
    return pool;
}




/* Destructor */
CfgPool cfgpool_delete (CfgPool pool) {

    ASSERT(pool);

    while (pool->stack) {
        struct CfgItem *item=pool->stack;

        /* Pop from the stack */
        pool->stack=pool->stack->nxt;
        
        /* Free current item */
        free(item->key);
        free(item->val);
        free(item);
    }
    
    free(pool);

    return NULL;
}






/*
    This function returns a dinamically created copy of the value
corresponding to the specified 'key', looking for it in the 'pool'
and converted to 'char *' type. Please note that this returned copy
must be freed by the caller...

    If the key is not found, (char *)'0' is returned. Just for symmetry
with the other 'get' methods, the caller shouldn't rely on the return
value for error checking. The 'errno' value must be checked instead.
Please note that the fact this function returns an invalid value in case
of error *in addition to the errno code* is just a coincidence.

    ESRCH: Cannot find the key.
    ENOMEM: Not enough memory to return the copy.
*/
char * cfgpool_getstring (CfgPool pool, const char *key) {

    struct CfgItem *item=NULL;

    ASSERT(pool);
    ASSERT(key);

    item=pool->stack;
    while (item) {
    
        if (!strcmp(item->key, key)) {
            char *val=NULL;
            
            val=malloc(strlen(item->val)+1);
            if (!val) {
                errno=ENOMEM;
                return NULL;
            }
            strcpy(val, item->val);  /* Copy the value */
            return val;
            break;
        }
        
        item=item->nxt;
    }   

    errno=ESRCH;
    return NULL;
}




/*
    This function returns the value corresponding to the specified
'key', looking for it in the 'pool' as a natural number, that is,
a number in the range [0,UINTMAX_MAX] with 'uintmax_t' type and
the specified 'base'. This number can converted to the appropriate
unsigned type afterwards, checking for the range, of course ;)

    In case of error (uintmax_t)'0' is returned, *which is a perfectly
legal value*. So, for proper error checking, the caller should zero
'errno' and check if it is zero after the call.

    ESRCH: Cannot find the key.
    ENOMEM: Not enough memory to perform the query.
    EINVAL: Couldn't parse a well-formed natural number.
    ERANGE: Overflow while parsing the number.
*/
uintmax_t
cfgpool_getnatural (CfgPool pool, const char *key, unsigned int base) {

    char *value=NULL;
    char *tailptr=NULL;
    int tmperrno=errno;
    uintmax_t natural=0;

    ASSERT(pool);
    ASSERT(key);
    ASSERT(base >= 2);  /* Better use a sane base ;) */
    ASSERT(base <= 36); /* Idem... */

    /* First, get the string value */
    errno=0;
    value=cfgpool_getstring(pool, key);
    if (errno) return 0;

    /* Well, it seems we have a value... Translate to correct type */
    errno=0;
    natural=strtoumax(value, &tailptr, base);
    
    if (errno) natural=0,tmperrno=ERANGE;
    if ((!natural) && (tailptr==value)) tmperrno=EINVAL;
    
    free(value);
    errno=tmperrno;
    return natural;
}




/*
    This function returns the value corresponding to the specified
'key', looking for it in the 'pool' as an integer number, that is,
a number in the range [INTMAX_MIN,INTMAX_MAX] with 'uintmax_t' type
and the specified 'base'. This number can converted to the appropriate
integer type afterwards, checking for the range, of course ;)

    In case of error (uintmax_t)'0' is returned, *which is a perfectly
legal value*. So, for proper error checking, the caller should zero
'errno' and check if it is zero after the call.

    ESRCH: Cannot find the key.
    ENOMEM: Not enough memory to perform the query.
    EINVAL: Couldn't parse a well-formed integer number.
    ERANGE: Overflow while parsing the number.
*/

intmax_t
cfgpool_getinteger (CfgPool pool, const char *key, unsigned int base) {

    char *value=NULL;
    char *tailptr=NULL;
    int tmperrno=errno;
    intmax_t integer=0;

    ASSERT(pool);
    ASSERT(key);
    ASSERT(base >= 2);  /* Better use a sane base ;) */
    ASSERT(base <= 36); /* Idem... */

    /* First, get the string value */
    errno=0;
    value=cfgpool_getstring(pool, key);
    if (errno) return 0;

    /* Well, it seems we have a value... Translate to correct type */
    errno=0;
    integer=strtoimax(value, &tailptr, base);
    
    if (errno) integer=0,tmperrno=ERANGE;
    if ((!integer) && (tailptr==value)) tmperrno=EINVAL;

    free(value);
    errno=tmperrno;
    return integer;
}
