/* *** chain.c *** The indescribable Chain container

Copyright (C) 2002,03 Ral Nez de Arenas Coronado

        This program is free software; you can redistribute it and/or
         modify it under the terms of the GNU General Public License
           version 2 as published by the Free Software Foundation.

       This program is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty
           of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
            See the GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
             (`GPL') along with this program; if not, write to:

                       Free Software Foundation, Inc.
                         59 Temple Place, Suite 330
                         Boston, MA 02111-1307  USA
*/
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <limits.h>
#include <stdint.h>
#include "chain.h"
#include "mobs.h"

/*
    NOTE: This dynamic sequential access data container is designed as a
circular doubly linked list, featuring a fake link. This fake link breaks
the list 'circularity', providing and start/end point (let's call it home)
for the container. Please note that this is entirely intentional. This fake
node is implemented as the Chain object itself, for memory saving purposes.

    A side (but intended) effect of this fake node is that the insertion of
the first element is NOT an special case, so insertions are a bit faster.
*/

/* Declared here for allowing forward references */
typedef struct Link *Link;


/*
    The two first fields make this struct compatible with 'struct Link'.
The third field may not be compatible with 'void *', but it doesn't matter,
since the 'data' field of this struct is *never* referenced.
*/
struct Chain {
    Link next;
    Link prev;
    Link cursor;
    size_t size;  /* Number of items in the chain */
    size_t offset;  /* Current index (offset) within the chain */
};


/* This struct defines a link into the chain */
struct Link {
    Link next;
    Link prev;
    void *data;
};



/*
    Constructor: builds a new Chain object and returns it.
    
    On error, it returns NULL and sets errno to ENOMEM.
    
    ENOMEM: Not enough memory to build the Chain object.
*/
Chain chain_create (void) {

    Chain chain=NULL;

    chain=malloc(sizeof(struct Chain));
    if (!chain) {
        errno=ENOMEM;
        return NULL;
    }

    /* Properly tie fake node (the chain itself)... */
    chain->next=(Link)chain;
    chain->prev=(Link)chain;
    chain->cursor=(Link)chain;
    
    chain->size=0;
    chain->offset=0;  /* When we are 'at home' we are at offset 0 */
    
    return chain;
}




/*
    Destructor: destroys the Chain object.

    Always returns 'NULL', for convenience: it should be called as
e.g. 'mychain=chain_delete(mychain)'. This way 'mychain' is not only
destroyed, but assigned 'NULL' so it doesn't get used by mistake in
any of the other methods.

    Please note that although the Chain doesn't own the data it contains,
if this destructor is called while the 'chain' is not empty, the nodes are
properly destroyed AND THE DATA CONTAINED IN THEM IS FREED using the 'free()'
library function. This is done because minimizes the risk of a memory leak.
*/
Chain chain_delete (Chain chain) {

    ASSERT(chain);

    /* Destroy all the nodes in the chain, WITHOUT freeing the data. */
    /* First go to 'home' */
    chain->cursor=(Link)chain;
    while(chain->size--) {
        Link target=chain->cursor->next;

        /* Free the item's data */
        free(target->data);
           
        /* Detach the node, no need to set up 'prev' link... */
        chain->cursor->next=target->next;
            
        /* Delete the node's data (helps debugging) */
        target->next=NULL;
        target->prev=NULL;
        target->data=NULL;
            
        /* Free the node's memory */
        free(target);
    }
        
    /* Let memory in a sane state. Helps debugging... */
    chain->next=NULL;
    chain->prev=NULL;
    chain->cursor=NULL;
    chain->size=0;
    chain->offset=0;
    free(chain);

    return NULL;
}




/*
    This returns the number of items in the 'chain'. On success, a number
between 0 and SIZE_MAX. No error conditions are assumed and neither the
'chain' nor its cursor are modified.

    This function is an accessor method.
*/
size_t chain_size (Chain chain) {

    ASSERT(chain);

    return chain->size;

}




/*
    This function tells the index offset in the 'chain'.
    An offset of '0' means the cursor is 'at home'.
    
    This function is an accessor method.
*/
size_t chain_offset (Chain chain) {

    ASSERT(chain);
    
    return chain->offset;
}




/*
    This function adds the 'item' to the 'chain', just after the cursor,
moving the cursor to the newly added item. This gives append semantics,
just like a file, but anywhere on the chain. Please note that no 'NULL'
objects can be added to a chain.

    On error, 'NULL' is returned. Otherwise, 'item' is returned.

    ENOMEM: Not enough memory to add object to the chain.
    EFBIG:  Chain is full of items (SIZE_MAX items already in the chain).
*/
void *chain_put (Chain chain, void *item) {

    Link newlink=NULL;

    ASSERT(chain);
    ASSERT(item);

    if (chain->size==SIZE_MAX) {
        errno=EFBIG;
        return NULL;
    }
    
    /* Create the new link */    
    newlink=malloc(sizeof(struct Link));
    if (!newlink) {
        errno=ENOMEM;
        return NULL;
    }

    /* Store the data into the newly created link */
    newlink->data=item;

    /* Link the new node to the list at the insertion point */    
    /* First link the node itself */
    newlink->prev=chain->cursor;        /* Point prev at cursor */
    newlink->next=newlink->prev->next;  /* Point next at cursor->next */

    /* Now link the next and prev links to this one */
    newlink->next->prev=newlink;
    newlink->prev->next=newlink;

    /* Displace the cursor */
    chain->cursor=newlink;
    
    /* One more to the chain... */
    chain->size++;
    chain->offset++;

    return item;
}




/*
    This function returns the data under the cursor.
    If the cursor is at home then 'NULL' is returned.
    The chain is not modified in any case.
*/
void *chain_get (Chain chain) {

    ASSERT(chain);

    return (chain->cursor==(Link)chain) ? NULL : chain->cursor->data;

}




/*
    This function removes the cursor item from 'chain', returning
a reference to the data contained in it, placing the cursor in the
preceeding item. If the cursor is at home then 'NULL' is returned
and the chain is not modified.
*/
void *chain_del (Chain chain) {

    Link target=NULL;
    void *data=NULL;

    ASSERT(chain);

    target=chain->cursor;  /* Target for detaching */
    if (target==(Link)chain) return NULL;  /* Cannot delete 'home' */

    data=target->data;
    chain->cursor=target->prev;

    /* Relink prev and next links to be able to dissapear */
    target->prev->next=target->next;
    target->next->prev=target->prev;

    free(target);

    /* One less brick in the wall */
    chain->size--;
    
    /* The offset is modified */
    chain->offset--;

    return data;  /* Previously saved */
}



/*
    This function moves the cursor to the next link in the chain.
    After moving, it returns the data under the cursor, or NULL if at home.
*/
void *chain_next (Chain chain) {

    ASSERT(chain);

    chain->cursor=chain->cursor->next;

    if (chain->cursor==(Link)chain) {
        chain->offset=0;
        return NULL;
    } else {
        chain->offset++;
        return chain->cursor->data;
    }

}




/*
    This function moves the cursor to the prev link in the chain.
    After moving, it returns the data under the cursor, or NULL if at home.
*/
void *chain_prev (Chain chain) {

    ASSERT(chain);

    chain->cursor=chain->cursor->prev;

    if (chain->cursor==(Link)chain) {
        chain->offset=0;
        return NULL;
    } else {
        chain->offset--;
        return chain->cursor->data;
    }
}




/*
    This function puts the chain cursor at 'home'. No more, no less.
*/
void chain_home (Chain chain) {

    ASSERT(chain);
    
    chain->cursor=(Link)chain;
    chain->offset=0;

}




/*
    This function performs random access on the 'chain'.
*/

void *chain_seek (Chain chain, size_t offset) {

    size_t nextdistance=0;
    size_t prevdistance=0;

    ASSERT(chain);

//    SAY("Entering");
//    SAY("Requested offset %d, current offset %d", offset, chain->offset);

    /* Well, if we are at home... */    
    if (!offset) {
        chain_home(chain);
        return NULL;
    }
    
    /* We cannot go beyond the end... */
    if (offset >= chain->size) {
        chain->cursor=chain->prev;
        chain->offset=chain->size;
        return chain->cursor->data;
    }
    
    if (offset == 1) {
        chain->cursor=chain->next;
        chain->offset=1;
        return chain->cursor->data;
    }

    /* Maybe we are already there... */
    if (offset == chain->offset) return chain->cursor->data;
    
    /* Determine shortest path... */
    if (offset < chain->offset) {
        prevdistance=chain->offset-offset;
        nextdistance=chain->size-prevdistance+1;
    } else {
        nextdistance=offset-chain->offset;
        prevdistance=chain->size-nextdistance+1;
    }

    if (nextdistance < prevdistance) {  /* nextwise is shorter */
        while (nextdistance--) chain->cursor=chain->cursor->next;
    } else {  /* prevwise is shorter */
        while (prevdistance--) chain->cursor=chain->cursor->prev;
    }

    chain->offset=offset;

    return chain->cursor->data;    

}
