/******************************************************************************
 *
 *  Filename: pafont.c
 *  Description: Generic font class
 *
 *                             Copyright (C) 2004
 *                 David Gmez Espinosa <david@pleyades.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include "pafont.h"
#include "mobs.h"

#ifdef DISABLE_XFT

/* Support for the old X font system that uses X unLogical Font Descriptors */
struct sPAFont {
	XFontStruct *font;
	GC gc;
	Display *display;
};

#define PAFontBytes (sizeof(struct sPAFont))



void pafont_init(void)
{
	SAY("Initializing pafont");
}


PAFont pafont_create(Display *dpy, const char* id)
{
	PAFont xlfdfont;
	char fontname[20];
	char *tmpvalue;

	ASSERT(id);

	xlfdfont=malloc(PAFontBytes);
	if (!xlfdfont) die("Not enough memory");

	xlfdfont->display=dpy;

	strcpy(fontname, id); strcat(fontname, "_xlfd");

	tmpvalue=cfgpool_getstring(pawmpool, fontname);
	if (tmpvalue==NULL) die(NOTFOUNDMSG,fontname);

	xlfdfont->font=XLoadQueryFont(xlfdfont->display, tmpvalue);
	if (!xlfdfont->font) BUG();

	xlfdfont->gc=XCreateGC(xlfdfont->display, ROOT, 0, NULL);
	if (xlfdfont->gc==None) BUG();

	XSetFont(xlfdfont->display, xlfdfont->gc, xlfdfont->font->fid);

	free(tmpvalue);

	return xlfdfont;
}


unsigned int pafont_width(PAFont xlfdfont, const char * text, size_t len)
{
	ASSERT(xlfdfont);

	return XTextWidth(xlfdfont->font, text, len);
}


void pafont_draw(PAFont xlfdfont, Window win, int x, int y, const char *text, size_t len)
{
	ASSERT(xlfdfont);

	/* X and Y coordinates mark the top left corner of the text box,
	   so we need to sum the ascent value to y to make the shitty old
	   X font system draw the text at the right place */
	XDrawString(xlfdfont->display, win, xlfdfont->gc, x, y+xlfdfont->font->ascent, text, len);
}


void pafont_destroy(PAFont xlfdfont)
{
	ASSERT(xlfdfont);

	XFreeGC(xlfdfont->display, xlfdfont->gc);
	XFreeFont(xlfdfont->display, xlfdfont->font);
}


/* Return the maximum height for given font */
unsigned int pafont_height(PAFont xlfdfont)
{
	ASSERT(xlfdfont);

	return (xlfdfont->font->ascent+xlfdfont->font->descent);
}


/* This function returns the number of full chars of 'text' which
will fit in the given width, 'w', if painted with 'font'.

NOTE: uses Bolzano's theorem to calculate the size.  */
size_t pafont_trimtext (PAFont xlfdfont, const char *text, unsigned int w)
{
	size_t min=0;
	size_t max=0;
	size_t pivot=0;

	ASSERT(xlfdfont);

	if (!text) return 0;

	max=strlen(text)+1;
	pivot=max/2;    

	/* We use Bolzano's theorem for calculate the text width */
	while (1) {
		unsigned int pivotw=0;

		/* Compute text length in pixels for 'pivot' chars */
		pivotw=XTextWidth(xlfdfont->font, text, (int)pivot);
		
		/* If the minimum is greater than the allowed width... */
		if (pivotw >= w) max=pivot;
		if (pivotw < w) min=pivot;
		if ((max-min) == 1) return min;  /* Algorithm converged */
		pivot=((max-min)/2)+min;
	}
}


void pafont_setcolor(PAFont xlfdfont, char *name)
{
	XColor ecolor, scolor;

	XAllocNamedColor(xlfdfont->display, COLORMAP, name, &scolor, &ecolor);
	XSetForeground(xlfdfont->display, xlfdfont->gc, scolor.pixel);
}

#else

#include <X11/Xft/Xft.h>

/* Xft font system */
struct sPAFont {
	XftFont *font;
	XftDraw *draw;
	XftColor *color;
	Display *display;
};

#define PAFontBytes (sizeof(struct sPAFont))



PAFont pafont_create(Display *dpy, const char* id)
{
	PAFont xftfont;
	char fontname[20];
	char *tmpvalue;

	ASSERT(id);

	xftfont=malloc(PAFontBytes);
	if (!xftfont) die("Not enough memory");

	xftfont->display=dpy;

	strcpy(fontname, id); strcat(fontname, "_xft");

	tmpvalue=cfgpool_getstring(pawmpool, fontname);
	if (tmpvalue==NULL) die(NOTFOUNDMSG,fontname);

	SAY("Opening font %s",tmpvalue);
	xftfont->font=XftFontOpenName(xftfont->display, SCREENI, tmpvalue);
	if (!xftfont->font) BUG();

	SAY("Creating XftDraw");
	xftfont->draw=XftDrawCreate(xftfont->display, ROOT, VISUAL, COLORMAP);
	if (!xftfont->draw) BUG();

	xftfont->color=malloc(sizeof(XftColor));

	free(tmpvalue);

	return xftfont;
}


unsigned int pafont_width(PAFont xftfont, const char * text, size_t len)
{
	XGlyphInfo extents;

	ASSERT(xftfont);

	XftTextExtentsUtf8(xftfont->display, xftfont->font, (XftChar8 *)text, len, &extents);   

	return extents.width;
}


void pafont_draw(PAFont xftfont, Window win, int x, int y, const char *text, size_t len)
{
	ASSERT(xftfont);

	/* X and Y coordinates mark the top left corner of the text box,
	   so we need to sum the ascent value to y to make the shitty old
	   X font system draw the text at the right place */

	/* TODO : Do we need to sum ascent? Ugly cast to Xftchar? */
	/* TODO: Check if the drawable is already selected */
	XftDrawChange(xftfont->draw, win);
	XftDrawStringUtf8(xftfont->draw, xftfont->color, xftfont->font,
		   			  x, y+xftfont->font->ascent, (XftChar8 *)text, len);
}


void pafont_destroy(PAFont xftfont)
{
	ASSERT(xftfont);

	SAY("Destroying font");
	XftDrawDestroy(xftfont->draw);
	XftFontClose(xftfont->display, xftfont->font);
}


/* Return the maximum height for given font */
unsigned int pafont_height(PAFont xftfont)
{
	ASSERT(xftfont);

	//SAY("ascent: %d descent %d height %d",xftfont->font->ascent,xftfont->font->descent,xftfont->font->height);
	return (xftfont->font->ascent+xftfont->font->descent);
}


/* This function returns the number of full chars of 'text' which
will fit in the given width, 'w', if painted with 'font'.

NOTE: uses Bolzano's theorem to calculate the size.  */
size_t pafont_trimtext (PAFont xftfont, const char *text, unsigned int w)
{
	size_t min=0;
	size_t max=0;
	size_t pivot=0;

	ASSERT(xftfont);

	if (!text) return 0;

	max=strlen(text)+1;
	pivot=max/2;    

	/* We use Bolzano's theorem for calculate the text width */
	while (1) {
		unsigned int pivotw=0;

		/* Compute text length in pixels for 'pivot' chars */
		pivotw=pafont_width(xftfont, text, (int)pivot);
		
		/* If the minimum is greater than the allowed width... */
		if (pivotw >= w) max=pivot;
		if (pivotw < w) min=pivot;
		if ((max-min) == 1) return min;  /* Algorithm converged */
		pivot=((max-min)/2)+min;
	}

}


void pafont_setcolor(PAFont xftfont, char *name)
{
	ASSERT(xftfont);

	XftColorAllocName(xftfont->display, VISUAL, COLORMAP, name, xftfont->color );
}
#endif
