/*
    Filename: pawindow.c
    Synopsis: PA managed Window.

                          Copyright (C) 2000-2006
            Raúl Núñez de Arenas Coronado <raul@pleyades.net>
	         David Gómez Espinosa <david@pleyades.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

*/

#include <stdbool.h>

#include "pawindow.h"
#include "palabel.h"
#include "pabar.h"
#include "pafont.h"
#include "paicon.h"
#include "hints.h"
#include "keyboard.h"


/*º Resource definitions: ***************************************************/

/*  - Cursors  */
#define NWCURSOR XC_ul_angle
#define NECURSOR XC_ur_angle
#define SWCURSOR XC_ll_angle
#define SECURSOR XC_lr_angle
#define RESIZECURSOR XC_sizing

/*  - States */
#define PA_ICONIFIED 	0
#define PA_MAPPED       1
#define PA_WITHDRAWN    2


/****************************************************************************/


/* Minimum border width, in pixels. */
#define BORDERW (2)

/* These are for easing calculations... */
/* And, don't worry: all pawindow vars are called 'pawindow' XDD */
#define TITLEBAR_X (BORDERW)
#define TITLEBAR_Y (BORDERW)
#define TITLEBAR_W (pawindow->w-BORDERW*2)
#define TITLEBAR_H 20
#define TITLEBAR_MARGIN 5
#define CANVAS_X (BORDERW)
#define CANVAS_Y (BORDERW+TITLEBAR_H)
#define CANVAS_W (pawindow->w-BORDERW*2)
#define CANVAS_H (pawindow->h-BORDERW*2-TITLEBAR_H)
#define SHADOW_X (pawindow->x+BORDERW)
#define SHADOW_Y (pawindow->y+BORDERW)
#define SHADOW_W (pawindow->w)
#define SHADOW_H (pawindow->h)

/* Absolute size limits */
#define FRAME_MIN_W (BORDERW*2+TITLEBAR_H+BUTTONS_W)
#define FRAME_MIN_H (BORDERW*2+TITLEBAR_H+1)
#define FRAME_MAX_W (SHRT_MAX-BORDERW*2)
#define FRAME_MAX_H (SHRT_MAX-BORDERW*2-TITLEBAR_H)

/* For the tab-window */
#define ICONW 48
#define ICONH 48	
#define TAB_MARGIN 5
#define TAB_FRAME_PADDING 3  // Padding to left space for the icons boxes
#define TAB_MAX_ICONS_PERROW 6	
#define TAB_W ((48*TAB_MAX_ICONS_PERROW)+(TAB_ICON_SPACING*(TAB_MAX_ICONS_PERROW-1))+(TAB_MARGIN*2)+(TAB_FRAME_PADDING*2))
#define TAB_H ((48*num_rows)+(TAB_ICON_SPACING*(num_rows-1))+pafont_height(class.font)+(TAB_MARGIN*2)+(TAB_FRAME_PADDING*2))
#define TAB_ICONW_W (TAB_W-(TAB_MARGIN*2))	
#define TAB_TEXTW_W (TAB_W-(TAB_MARGIN*2))	
#define TAB_ICON_SPACING 6

/* Event masks for the pawindow parts */
#define CANVAS_EVMASK (SubstructureRedirectMask | SubstructureNotifyMask| ButtonPressMask | FocusChangeMask | KeyPressMask | KeyReleaseMask)
#define FRAME_EVMASK (ButtonPressMask | ButtonReleaseMask | ExposureMask | EnterWindowMask | LeaveWindowMask)
#define TITLEBAR_EVMASK (ButtonPressMask | ButtonReleaseMask | ExposureMask)

/*º Data Structures *********************************************************/

/* The tab-window */
struct TabW {
	Window tab;
	Window icons;
	Window text;
};

/* This struct contains the per-class data */
static struct {
	PAFont font;
	GC gc;
	struct {
		unsigned long passivebar;
		unsigned long activebar;
		char * passivetext;
		char * activetext;
		char * textshadow;
		unsigned long framelight;
		unsigned long framewhite;
		unsigned long framebase;
		unsigned long framedark;
		unsigned long framedarker;
	} color;
	struct {
		Cursor NW;
		Cursor NE;
		Cursor SW;
		Cursor SE;
		Cursor resize;
	Cursor move;
	} cursor;
	struct {
		xImage close_normal;
		xImage close_active;
		xImage close_pressed;
		xImage iconz_normal;
		xImage iconz_active;
		xImage iconz_pressed;
		xImage maximize_normal;
		xImage maximize_active;
		xImage maximize_pressed;
	} face;  /* Button faces */
	struct {
		Button close;
		Button iconz;
		Button maximize;
	} button;  /* Buttons */
	Bool shadow_status;	          /* shadow status */
	unsigned int titlebar_height; /* Titlebar height */
	unsigned int texty;           /* Where titlebar text is written */
	Time timestamp;               /* Used to check for double-click */
	int clickx, clicky;          
	Chain pawindows;              /* Tracks pawindows */
	Chain fpawindows;             /* Pawindows that have had the focus */
	struct TabW tab_window;
	XTransform scale_matrix;
} class;


/* This is the object itself */
struct sPAWindow {
	Window window;       /* Here we put the client window */
	Window shadow;       /* Shadow on the wall... */
	Window frame;        /* Frame around the client */
	Window canvas;       /* Client canvas */
	Window titlebar;     /* Titlebar, of course... */
	Window wmicon_win;   /* Window for the tab icon*/
	xImage wmicon;       /* Image for the tab-icon*/
	Picture wmpicture;   /* Render picture for the tab-icon */
	Bool active;         /* Has this pawindow the focus? */
	Bool transient;      /* Is this pawindow transient? */
	Bool decoration;     /* Does it need decoration? */
	Bool shaded;         /* Is this pawindow shaded? */
	Bool shaped;         /* Does it have a shape */
	Bool dialog;         /* Dialog kind of window: _NET_WM_WINDOW_TYPE_DIALOG */
	Bool modal;			 /* Modal window: _NET_WM_STATE_MODAL */		
	Bool utility;		 /* Utility window: _NET_WM_WINDOW_TYPE_UTILITY*/
	PAWindow modalchild;  /* Modal childs */
	Bool fullscreen;     /* Using _NET_WM_NAME_FULLSCREEN */
	Bool got_netwmname;  /* _NET_WM_NAME property */
	Bool got_netwmiconname;  /* _NET_WM_ICON_NAME property */
	Bool got_skiptaskbar;    /* _NET_WM_SKIP_TASKBAR property */
	int x, y;            /* Position of a pawindow */
	unsigned int w, h;   /* Dimension of a pawindow */
	unsigned int min_w;  /* Minimum width of client */
	unsigned int min_h;  /* Minimum height of client */
	unsigned int max_w;  /* Maximum width of client */
	unsigned int max_h;  /* Maximum height of client */
	unsigned int w_inc;  /* Width increment*/
	unsigned int h_inc;  /* Height increment */
	PALabel icon;        /* Iconized form of a pawindow */
	unsigned long state; /* State of a pawindow */
	char *title;         /* Text appearing onto titlebar */
	int oldx, oldy;
	unsigned int oldw, oldh; /* Remember previous geometry , used
								for un(clefucka)maximizing a pawindow */
	int oldfx, oldfy;
	unsigned int oldfw, oldfh; /* Remember previous geometry for fullscreening */
	int oldsh;           /* Remember previous height , used 
							for unshading a pawindow */ 
    
};

#define PAWindowBytes sizeof(struct sPAWindow)

/****************************************************************************/




/*º Globals *****************************************************************/

PAWindow focuspawindow=NULL;
static unsigned short int reference_count=0;

/****************************************************************************/




/*º Function prototypes *****************************************************/

/* Handlers: CANVAS */
static void canvas_handle (XEvent *, void *);
static void canvas_DestroyNotify    (XDestroyWindowEvent *, PAWindow);
static void canvas_UnmapNotify      (XUnmapEvent *, PAWindow);
static void canvas_ConfigureRequest (XConfigureRequestEvent *, PAWindow);
static void canvas_PropertyNotify   (XPropertyEvent *, PAWindow);
static void canvas_MapRequest       (XMapEvent *, PAWindow);
static void canvas_ClientMessage    (XClientMessageEvent *, PAWindow);
static void canvas_ClientMessage_netwmstate (XClientMessageEvent *, PAWindow);
static void canvas_FocusOut         (XFocusOutEvent *, PAWindow);
static void canvas_ReparentNotify   (XReparentEvent *, PAWindow);
static void canvas_KeyPress   (XKeyPressedEvent *, PAWindow);

/* Handlers: FRAME */
static void frame_handle (XEvent *, void *);
static void frame_ButtonPress (XButtonPressedEvent *, PAWindow);
static void frame_Expose      (XExposeEvent *, PAWindow);
static void frame_EnterNotify (XEnterWindowEvent *, PAWindow);
static void frame_LeaveNotify (XEnterWindowEvent *, PAWindow);

/* Handlers: TITLEBAR */
static void titlebar_handle (XEvent *, void *);
static void titlebar_ButtonPress (XButtonPressedEvent *, PAWindow);
static void titlebar_Expose      (XExposeEvent *, PAWindow);

/* Drawing functions */
static void draw_titlebar (PAWindow);
static void draw_border   (PAWindow);
static void draw_outline  (int, int, int, int);

/* Resize and move helpers */
static void take_a_trip        (PAWindow);
static void stretch_and_shrink (PAWindow);

/* Other helpers */
//static void kill_client (void *);
//static void iconify_client (void *);
//static void maximize_client (void *);
static void restore_client (void *);
static void set_title (PAWindow, char *, size_t);
static void set_shape (PAWindow);
static void set_client_list(void);
static void set_fullscreen(PAWindow, bool);

static void pawindow_get_state(PAWindow);
static void pawindow_get_type(PAWindow);

/****************************************************************************/


/* Inline functions for using the chain with stack semantics */
inline void chain_push(Chain chain, void *object)
{
	PAWindow pw;

	ASSERT(chain);
	ASSERT(object);

	pw= (PAWindow) object;

	chain_home(chain);
	chain_put(chain, pw);
}


inline void *chain_pop(Chain chain)
{
	PAWindow pw;

	ASSERT(chain);

	chain_seek(chain,1);
	pw=chain_del(chain);

	return pw;
}


/****************************************************************************/
/*                                                                          */
/*º                        Event handling routines                          */
/*                                                                          */
/****************************************************************************/

/*º Canvas event handling ***************************************************/
static void canvas_handle (XEvent *xevent, void *widget)
{
	PAWindow pawindow=NULL;
	int type, foo;

	ASSERT(xevent);
	ASSERT(widget);

	pawindow=(PAWindow)widget;

	/* Remap pawindow we receive a shape event, just to make visible
	   the changes */
	XShapeQueryExtension(display,&type,&foo);
	if ( xevent->type==type) {
		GOT(ShapeNotify,Canvas);
		return;
	}

	switch (xevent->type) {
		case DestroyNotify:
			GOT(DestroyNotify, Canvas);
			canvas_DestroyNotify(&(xevent->xdestroywindow), pawindow);
			break;
		case UnmapNotify:
			GOT(UnmapNotify, Canvas);
			canvas_UnmapNotify(&(xevent->xunmap), pawindow);
			break;
		case ConfigureRequest:
			GOT(ConfigureRequest, Canvas);
			canvas_ConfigureRequest(&(xevent->xconfigurerequest), pawindow);
			break;
		case PropertyNotify:
			GOT(PropertyNotify, Canvas);
			canvas_PropertyNotify(&(xevent->xproperty), pawindow);
			break;
		case ButtonPress:
			GOT(ButtonPress, Canvas); /* Just ignore it to avoid propagate to frame*/	
			break;
		case MapRequest:
			GOT(MapRequest, Canvas);
			canvas_MapRequest(&(xevent->xmap), pawindow);
			break;
		case ClientMessage:
			GOT(ClientMessage, Canvas);
			canvas_ClientMessage(&(xevent->xclient), pawindow);
			break;
		case FocusIn:
			GOT(FocusIn, Canvas);
			break;
		case FocusOut:
			GOT(FocusOut, Canvas);
			/* This breaks gucharmap */
			//canvas_FocusOut(&(xevent->xfocus), pawindow);
			break;
		case ReparentNotify:
			GOT(ReparentNotify, Canvas);
			canvas_ReparentNotify(&(xevent->xreparent), pawindow);
			break;
		case KeyPress:
			GOT(KeyPress, Canvas);
			canvas_KeyPress(&(xevent->xkey), pawindow);
			break;
		case KeyRelease:
			GOT(KeyRelease, Canvas);
			break;
		case MapNotify:
			GOT(MapNotify, Canvas);
			break;	
		default:
			SAY("Event type %d not handled on Canvas",xevent->type);
			break;
	}
}



static void canvas_DestroyNotify (XDestroyWindowEvent *event,
                                  PAWindow pawindow)
{

	ASSERT(event);
	ASSERT(pawindow);    
	if (event->window!=pawindow->window) BUG();

	pawindow_destroy(pawindow);

}




static void canvas_UnmapNotify (XUnmapEvent *event, PAWindow pawindow)
{
	XEvent xevent;

	ASSERT(event);
	ASSERT(pawindow);
	if (event->window!=pawindow->window) BUG();
	 
	/* Ignore UnmapNotify coming from pawindows being shaded, but
	   keeping and eye on the window being destroyed */
	if (pawindow->shaded) return;

	pawindow_unmap(pawindow);

	/* See if the client wants to go Withdrawn */
	/*
	    As per ICCCM, when a ICCCM-aware client wants to withdraw one
	of its windows it should unmap the window and, after that, send a
	synthetic UnmapNotify to the root window. This is quite congruent,
	and the proper way of dealing with this protocol is to handle the
	event standing in one leg, with the arms on the back, telling three
	times 'Klaatu verata nictu" (reversed), casting a fireball spell on
	the XServer and invoking the Ancient Ones while you do the XRequest...
	*/
	if (XCheckTypedWindowEvent(display, ROOT, UnmapNotify, &xevent)==True) {

		CARD32 data[2];  

		SAY("Pawindow %#lx going withdrawn", pawindow->window);
		pawindow->state=PA_WITHDRAWN;
		pabar_dellabel(pawindow->icon);
		data[0]=WithdrawnState;
		data[1]=None;
		x_set_property(pawindow->window, atoms.wm_state, XA_CARDINAL, (unsigned char *)&data, 2);
	}

}



static void canvas_ConfigureRequest (XConfigureRequestEvent *event, PAWindow pawindow)
{
    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();

	if (!pawindow->transient && !pawindow->dialog) { 
		/* Get only useful values (specified in the value_mask) */
		if (event->value_mask & CWX || event->value_mask & CWY) {
			SAY("Requested x: %d",event->x);
			SAY("Requested y: %d",event->y);
			pawindow_move(pawindow, event->x, event->y);
		}
	}   /* Else: Fuck off the defaults, we'll center the pawindows */
    
    /* Make sure the client fits on the display when it's configuring itself */
    if (event->value_mask & CWWidth || event->value_mask & CWHeight) {

		SAY("Requested width: %d",event->width);
		SAY("Requested height: %d",event->height);
		
		/* Sanitize width */
		if ( event->width > DWIDTH ) {
			event->width=DWIDTH-BORDERW*2;
			event->x=BORDERW;
		}
		
		/* Sanitize height */
		if ( event->height > DHEIGHT ) {
			event->height=DHEIGHT-BORDERW*2-TITLEBAR_H;
			event->y=0;
		}
		
		pawindow_resize(pawindow, event->width, event->height);
		/* FIXME: Implement a pawindow_center function */
		pawindow_map(pawindow);
	
	
    } else {

        /* Some windows in withdrawn state generate a ConfigureRequest event,
           but no MapRequest, so it's necessary to map them or they'll never
           become visible. Welcome to The X Horror Show */
		XGrabServer(display);
		if (pawindow!=focuspawindow) {
    		pawindow_map(pawindow);
		}    
        XUngrabServer(display);
    }
    
    /* Shape windows doesn't generate MapRequests when are reshaped *sigh*, so it
       neccesary to redraw them at ConfigureRequest events */
    if (pawindow->shaped) 
        set_shape(pawindow);
    
    XFlush(display);
    
}




static void canvas_PropertyNotify (XPropertyEvent *event, PAWindow pawindow)
{
	char *atomname;
	unsigned long length;
	unsigned char *text;
	
	ASSERT(event);
	ASSERT(pawindow);
	
	if (event->window!=pawindow->window) BUG();

	atomname=XGetAtomName(display,event->atom);
	SAY("Atom received : %s", atomname);

	if (event->atom==atoms.net_wm_name) {
		SAY("Changing pawindow title using _NET_WM_NAME");
		text=x_get_property(pawindow->window, atoms.net_wm_name,
						atoms.utf8_string, &length);
		if (!text) return;
		
		pawindow->got_netwmname=True;
		set_title(pawindow, (char *)text, length);
		XFree(text);
	
	} else if (event->atom==atoms.wm_name) {	
		/* Ignore it if we already got a _NET_WM_NAME */
		if (!pawindow->got_netwmname) {
			SAY("Changing pawindow title using WM_NAME");
			text=x_get_property(pawindow->window, atoms.wm_name, atoms.string, &length);
			if (!text) return;

			set_title(pawindow, (char *)text, length);
			palabel_settext(pawindow->icon, (char *)text, length);
			XFree(text);
		}    
	} else if (event->atom==atoms.net_wm_icon_name) {	
		SAY("Changing icon title using _NET_WM_ICON_NAME");
		text=x_get_property(pawindow->window, atoms.net_wm_icon_name,
						atoms.utf8_string, &length);
		if (!text) return;
		
		pawindow->got_netwmiconname=True;
		palabel_settext(pawindow->icon, (char *)text, length);
		XFree(text);
	} else if (event->atom==atoms.wm_icon_name) {	
		/* Ignore it if we already got a _NET_WM_ICON_NAME */
		if (!pawindow->got_netwmname) {
			SAY("Changing icon title WM_ICON_NAME");
			text=x_get_property(pawindow->window, atoms.wm_icon_name,
							atoms.string, &length);
			if (!text) return;
			
			palabel_settext(pawindow->icon, (char *)text, length);
			XFree(text);
		}
	} else if (event->atom==atoms.wm_normal_hints) {	
        	XSizeHints *hints=NULL;
        	long mask=0;

        	hints=XAllocSizeHints();
        	if (!hints) return;
        	if (!XGetWMNormalHints(display, event->window, hints, &mask)) return;

		/* Get size limits. Ignoring mask value and using instead flags 
		 * defined in XSizeHints */
		if (hints->flags & PMinSize) {
			pawindow->min_w=hints->min_width;
			pawindow->min_h=hints->min_height;

			/* Only sanitize values when the pawindow is decorated, cause
			 * non-decorated windows are allowed to be as small as they want */
			if (pawindow->decoration) {
				if (pawindow->min_w < FRAME_MIN_W) pawindow->min_w=FRAME_MIN_W;
				if (pawindow->min_h < FRAME_MIN_H) pawindow->min_h=FRAME_MIN_H;
				if (pawindow->min_w > FRAME_MAX_W) pawindow->min_w=FRAME_MAX_W;
				if (pawindow->min_h > FRAME_MAX_H) pawindow->min_h=FRAME_MAX_H;
			}
	    
		}

		if (hints->flags & PMaxSize) {
		    pawindow->max_w=hints->max_width;
		    pawindow->max_h=hints->max_height;

		    if (pawindow->max_w==0) pawindow->max_w=FRAME_MAX_W;
		    if (pawindow->max_h==0) pawindow->max_h=FRAME_MAX_H;
		    if (pawindow->max_w < FRAME_MIN_W) pawindow->max_w=FRAME_MIN_W;
		    if (pawindow->max_h < FRAME_MIN_H) pawindow->max_h=FRAME_MIN_H;
		    if (pawindow->max_w > FRAME_MAX_W) pawindow->max_w=FRAME_MAX_W;
		    if (pawindow->max_h > FRAME_MAX_H) pawindow->max_h=FRAME_MAX_H;
		}

		/* Adjust values. Some clients have a bad attitude. */
		if (pawindow->max_w < pawindow->min_w)
		    pawindow->max_w=pawindow->min_w;
		if (pawindow->max_h < pawindow->min_h)
		    pawindow->max_h=pawindow->min_h;

		XFree(hints);
	} else if (event->atom==atoms.wm_transient_for) {	
		Window parent=None; PAWindow pa_parent=NULL;
		
		/* Center window relative to parent */
		XGetTransientForHint(display, pawindow->window, &parent);
		pawindow->transient=(parent==None) ? False : True;
		chain_home(class.pawindows);
		while ((pa_parent=chain_next(class.pawindows)) != NULL) {
			if (pa_parent->window==parent) {
				if (pawindow->w < pa_parent->w) {
					pawindow->x=((pa_parent->w - pawindow->w -(BORDERW*2))/2) + pa_parent->x;
					pawindow->y=((pa_parent->h - pawindow->h -(TITLEBAR_H+(BORDERW*2)))/2) + pa_parent->y;
				} else {
					pawindow->x= pa_parent->x - ((pawindow->w-pa_parent->w)/2);
					pawindow->y=((pa_parent->h - pawindow->h -(TITLEBAR_H+(BORDERW*2)))/2) + pa_parent->y;
				}
			}
		}
	} else if (event->atom==atoms.net_wm_state) {	
	
		/* Get the client state */
		pawindow_get_state(pawindow);
		
		/* Update the pawindow after recovering its state */
		pawindow_map(pawindow);
		
		if (pawindow->got_skiptaskbar) {
			pabar_dellabel(pawindow->icon);
		}	
	} else if (event->atom==atoms.motif_wm_hints) {	
		Window tmpwindow; 
		
		tmpwindow=pawindow->window;
		pawindow_destroy(pawindow);
		pawindow=pawindow_create(tmpwindow);
		pawindow->decoration=gtkHints_needDecoration(pawindow->window);
		pawindow_map(pawindow);
	}	

}


static void canvas_MapRequest(XMapEvent *event, PAWindow pawindow)
{
	
	ASSERT(event);
	ASSERT(pawindow);
	if (event->window!=pawindow->window) BUG();

	/* First time a window is created, MapRequest arrives at root window,
	 * and any other MapRequest will be send to the canvas (cause is his parent 
	 * window and has SubstructureRedirectMask selected) */

	XGrabServer(display);
	pawindow_map(pawindow);
	XUngrabServer(display);
    
}


static void canvas_ClientMessage(XClientMessageEvent *event, PAWindow pawindow)
{

	ASSERT(event);
	ASSERT(pawindow);
	if (event->window!=pawindow->window) BUG();

	//SAY("CLIENT MESSAGE TYPE %d",event->message_type);
	if (event->message_type==atoms.net_wm_state) {
		canvas_ClientMessage_netwmstate(event, pawindow);
	} else if (event->message_type==atoms.wm_change_state) {     
		if (event->data.l[0]==IconicState)
			pawindow_unmap(pawindow);
	}

}


/* This event is used to change the state for a _mapped_ window */
static void canvas_ClientMessage_netwmstate (XClientMessageEvent *event, PAWindow pawindow)
{
	SAY("Got _NET_WM_STATE on %#lx",pawindow->window);
	
	switch (event->data.l[0]) { 
		case 0: /* Remove */
			if(event->data.l[1]==atoms.net_wm_state_modal) {
				SAY("_NET_WM_STATE_MODAL removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_sticky) {
				SAY("_NET_WM_STATE_STICKY removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_maximized) {
				SAY("_NET_WM_STATE_MAXIMIZED removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_shaded) {
				SAY("_NET_WM_STATE_SHADED removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_skip_taskbar) {
				SAY("_NET_WM_SKIP_TASKBAR removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_skip_pager) {
				SAY("_NET_WM_STATE_SKIP_PAGER removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_hidden) {
				SAY("_NET_WM_STATE_HIDDEN removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_fullscreen) {
				SAY("_NET_WM_STATE_FULLSCREEN removed on %#lx",pawindow->window);
				pawindow->fullscreen=false;
				pawindow_map(pawindow);
			}
			else if(event->data.l[1]==atoms.net_wm_state_above) {
				SAY("_NET_WM_STATE_ABOVE removed on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_below) {
				SAY("_NET_WM_STATE_BELOW removed on %#lx",pawindow->window);
			}
			break; 
		case 1: /* Add */
			if(event->data.l[1]==atoms.net_wm_state_modal) {
				SAY("_NET_WM_STATE_MODAL added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_sticky) {
				SAY("_NET_WM_STATE_STICKY added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_maximized) {
				SAY("_NET_WM_STATE_MAXIMIZED added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_shaded) {
				SAY("_NET_WM_STATE_SHADED added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_skip_taskbar) {
				SAY("_NET_WM_SKIP_TASKBAR added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_skip_pager) {
				SAY("_NET_WM_STATE_SKIP_PAGER added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_hidden) {
				SAY("_NET_WM_STATE_HIDDEN added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_fullscreen) {
				SAY("_NET_WM_STATE_FULLSCREEN added on %#lx",pawindow->window);
				/* Make the pawindow appear as fullscreen, saving the old geometry */
				pawindow->fullscreen=true;
				pawindow_map(pawindow);
			}
			else if(event->data.l[1]==atoms.net_wm_state_above) {
				SAY("_NET_WM_STATE_ABOVE added on %#lx",pawindow->window);
			}
			else if(event->data.l[1]==atoms.net_wm_state_below) {
				SAY("_NET_WM_STATE_BELOW added on %#lx",pawindow->window);
			}
			break; 
	}    
}

#if 0
static void canvas_FocusOut(XFocusOutEvent *event, PAWindow pawindow)
{
    
    ASSERT(event);
    ASSERT(pawindow);
    
    /* Some toolkits like the fucking XITK steal the focus, some we have
       to take it back again to pawm */
    if (event->detail==NotifyNonlinear || event->detail==NotifyVirtual) {
        if (event->window!=pawindow->window) {SAY("focus out not in pawindow");}
        if (focuspawindow) {
	    SAY("taking back focus");
            XSetInputFocus(display,focuspawindow->window,
	  	           RevertToPointerRoot,CurrentTime); }
    }
}
#endif


static void canvas_ReparentNotify(XReparentEvent *event, PAWindow pawindow)
{
    
	ASSERT(event);
	ASSERT(pawindow);

	if (event->parent!=pawindow->canvas) {
		/* Someone is messing with our windows (you bastard! ;)) and
		   reparenting them. So we destroy the previous pawindow to
	   	   avoid "limbo" pawindows ;). Typically someone means the
		   fucking Qt library putting windows in fullscreen mode :( */
		pawindow_destroy(pawindow);
	}
}


static void canvas_KeyPress(XKeyPressedEvent *event, PAWindow pawindow)
{
	ASSERT(event);
	ASSERT(pawindow);

	XEvent *e = (XEvent *) event;

	/* Replay the event if it isn't a pawm keybinding */
	if (!keyboard_binding_key(event)) {
		if (!XSendEvent(display, pawindow->window, False, KeyPressMask, e)) BUG();
	}

	SAY("Got %d keycode with state %d", event->keycode, event->state);

	keyboard_gotkey(event->keycode, event->state);
	
}

/****************************************************************************/


/*º Frame event handling ****************************************************/
static void frame_handle (XEvent *xevent, void *widget)
{
	PAWindow pawindow=(PAWindow)widget;

	ASSERT(xevent);
	ASSERT(pawindow);

	switch (xevent->type) {
	case ButtonPress:
	    GOT(ButtonPress, Frame);
	    frame_ButtonPress(&(xevent->xbutton), pawindow);
	    break;
	case Expose:
	    GOT(Expose, Frame);
	    frame_Expose(&(xevent->xexpose), pawindow);
	    break;
	case EnterNotify:
	    GOT(EnterNotify, Frame);
	    frame_EnterNotify(&(xevent->xcrossing), pawindow);
	    break;
	case LeaveNotify:
	    GOT(LeaveNotify, Frame);
	    frame_LeaveNotify(&(xevent->xcrossing), pawindow);
	    break;
	default:
	    SAY("Event %d on Frame",xevent->type);
	    break;	

	}
}




static void frame_ButtonPress (XButtonPressedEvent *event, PAWindow pawindow)
{
	ASSERT(event);
	ASSERT(pawindow);    

	if (event->button!=Button1) return;
	if (event->window!=pawindow->frame) return;

	/* A click on the frame border causes a resize, but a click on
	   the window (frame child) causes a click-to focus policy */
	if ( ((event->x < BORDERW) ||                           // Left border   
	     (event->y < BORDERW) ||                            // Top border    
	     (event->x >= (int)(pawindow->w-BORDERW)) ||        // Right border  
	     (event->y >= (int)(pawindow->h-BORDERW)) ) &&   // Bottom border 
	     pawindow->decoration) {               
	    
		XGrabServer(display);
		pawindow_map(pawindow);
		XUngrabServer(display);

		/* Resize the pawindow */
		stretch_and_shrink(pawindow);
		/* And make sure that the pointer is ungrabbed with shaded windows */
		if (pawindow->shaded) 
			XUngrabPointer(display, CurrentTime);
	} else {
	    
		/* Avoid unnecesary work */
		if (pawindow==focuspawindow) { XAllowEvents(display, ReplayPointer, CurrentTime); return; }
		/* Click-to-focus :)) */
		//XGrabServer(display);
		pawindow_map(pawindow);
		//XUngrabServer(display);

		/* This must be done in the last place, or bad things can
		   happen (i.e. ButtonRelease events are not reported to the
	           window where the ButtonPress was delivered) */
		XAllowEvents(display, ReplayPointer, CurrentTime);
	    
	}

}


static void frame_Expose (XExposeEvent *event, PAWindow pawindow) 
{
	ASSERT(event);
	ASSERT(pawindow);

	/* Ugly return for decorated windows, could it be avoided?*/
	if (event->window!=pawindow->frame) return;
	    
	/* More Expose events to come */
	//if (event->count) return;

	/* Update the frame art ;) */
	draw_border(pawindow);
}



static void frame_EnterNotify (XEnterWindowEvent *event, PAWindow pawindow)
{
	ASSERT(event);
	ASSERT(pawindow);

	/* Ditto, see frame_Expose... */
	if (event->window!=pawindow->frame) return;

	XGrabButton(display, Button1, AnyModifier, pawindow->frame, False,
			ButtonPressMask, GrabModeSync, GrabModeSync, None, None);
}

static void frame_LeaveNotify(XEnterWindowEvent *event, PAWindow pawindow)
{
	ASSERT(event);
	ASSERT(pawindow);

	/* Sigh... X S-U-C-K-S */
	if (event->window!=pawindow->frame) return;

	/* Ungrab? No thanks XDD. The passive grab created by XGrabButton
	   changes to an active grab, which is automagically released when
	   a ButtonRelease event is received */
}
    

/****************************************************************************/



/*º Titlebar event handling *************************************************/
static void titlebar_handle (XEvent *xevent, void *widget) 
{
	PAWindow pawindow=NULL;

	ASSERT(xevent);
	ASSERT(widget);

	pawindow=(PAWindow)widget;

	switch (xevent->type) {
		case ButtonPress:
			GOT(ButtonPress, Titlebar);
			titlebar_ButtonPress(&(xevent->xbutton), pawindow);
			break;
		case Expose:
			GOT(Expose, Titlebar);
			titlebar_Expose(&(xevent->xexpose), pawindow);
			break;
		default:
			SAY("Event %d on Titlebar",xevent->type);
			break;	
	}
}




static void titlebar_ButtonPress (XButtonPressedEvent *event, PAWindow pawindow) 
{
	ASSERT(event);
	ASSERT(pawindow);

	if (event->window!=pawindow->titlebar) BUG();
	if (event->button!=Button1) return;

	XGrabServer(display);
	pawindow_map(pawindow);
	XUngrabServer(display);

	/* These thresholds should probably be configurable */
	if ( (labs(event->time-class.timestamp) < 400) &&
	     (abs(event->x_root-class.clickx) < 5) && 
	     (abs(event->y_root-class.clicky) < 5) ) 
	{
		if (pawindow->shaded) 
			pawindow_unshade(pawindow); /* Unshade the window */
		else 
			pawindow_shade(pawindow);   /* Shade this fucking window */
		class.timestamp=0;
		class.clickx=-1;
		class.clicky=-1;
	    
	} else {
		class.timestamp=event->time;
		class.clickx=event->x_root;
		class.clicky=event->y_root;
		/* Make a fucking travel :))) */
		take_a_trip(pawindow);
	}
}




static void titlebar_Expose (XExposeEvent *event, PAWindow pawindow)
{
	ASSERT(event);
	ASSERT(pawindow);
	if (event->window!=pawindow->titlebar) return;

	/* More Expose events to come. */
	if (event->count) return;

	/* Update the titlebar art ;) */
	draw_titlebar(pawindow);

}



/****************************************************************************/



/****************************************************************************/
/*                                                                          */
/*º                      Private (static) functions                         */
/*                                                                          */
/****************************************************************************/



/*º Drawing functions *******************************************************/

/* NOTE: ANY of the draw_*() functions can modify the GC. Be careful... */

/* This one redraws the title, if any */
static void draw_titlebar (PAWindow pawindow) 
{
	unsigned int w=TITLEBAR_W;  /* Maximum width available for title */
	int x=0;
	size_t len=0;

	ASSERT(pawindow);

	if (!pawindow->decoration) return;

	/* Draw politely */
	XSetFunction(display, class.gc, GXcopy);
	//XSetSubwindowMode(display, class.gc, ClipByChildren);

	/* Active windows have different decorations */
	if (pawindow->active==True)
		XSetWindowBackground(display, pawindow->titlebar, class.color.activebar);
	else
		XSetWindowBackground(display, pawindow->titlebar, class.color.passivebar);

	XClearWindow(display, pawindow->titlebar);
	if (!(pawindow->title)) return;

	/* Compute text length allowed for avoiding cut-text */
	w-=(2*TITLEBAR_MARGIN)+(BUTTONSZ*3);  /* Leave a pretty margin */

	len=pafont_trimtext(class.font, pawindow->title, w);

	/* Now center the title */
	/* 'w' contains the available space, and 'len' the text length */
	x= (w-pafont_width(class.font, pawindow->title, len))/2;

	/* Special decorations for active/passive windows */
	if (pawindow->active==True) {
		/* The text has a shadow on active windows */
		pafont_setcolor(class.font, class.color.textshadow);
		pafont_draw(class.font, pawindow->titlebar, x+2, class.texty+2, pawindow->title, len);
		pafont_setcolor(class.font, class.color.activetext);
	} else
		pafont_setcolor(class.font, class.color.passivetext);

	/* Finally add the text */
	pafont_draw(class.font, pawindow->titlebar, x, class.texty, pawindow->title, len);
}




/* Draw the Puto Amo(tm) 3D border around the Puto Amo(tm) window */
static void draw_border (PAWindow pawindow)
{
	unsigned int w=pawindow->w;
	unsigned int h=pawindow->h;

	if (!pawindow->decoration) return;

	XSetFunction(display, class.gc, GXcopy);

	/*
	    NOTE: the amazingly stupid X Window Fucking System has a very
	useful feature: If you paint a line from left to right the last point
	is not painted (is an X convention, I suppose), but if you paint the
	very same line from right to left, the first point misses :))))

	    This way, you end up with absurd endpoint specs and being *very*
	careful about how you draw a line :) The X Window System is quite useful,
	no doubt...
	*/

	/* NorthWest */
	/* 0 */
	XSetForeground(display, class.gc, class.color.framelight);
	XDrawLine(display, pawindow->frame, class.gc, 0, 0,(int) w, 0);
	XDrawLine(display, pawindow->frame, class.gc, 0, 0, 0,(int) h);

	/* 1 */
	XSetForeground(display, class.gc, class.color.framedark);
	XDrawLine(display, pawindow->frame, class.gc, 1, 1,(int) w-1, 1);
	XDrawLine(display, pawindow->frame, class.gc, 1, 1, 1,(int) h-1);

	/* SouthEast */
	/* 0 */
	XSetForeground(display, class.gc, class.color.framedarker);
	XDrawLine(display, pawindow->frame, class.gc, 0,(int) h-1,(int)w,(int)h-1);
	XDrawLine(display, pawindow->frame, class.gc,(int) w-1, 0,(int)w-1,(int)h);

	/* 1 */ /* 2 */
	XSetForeground(display, class.gc, class.color.framewhite);
	XDrawLine(display, pawindow->frame, class.gc, 1,(int) h-2,(int)w-2,(int)h-2);
	XDrawLine(display, pawindow->frame, class.gc,(int) w-2, 1,(int)w-2,(int)h-2);

}




static void draw_outline (int x1, int y1, int x2, int y2) 
{
	unsigned int w=(x2-x1);
	unsigned int h=(y2-y1);

	XSetSubwindowMode(display, class.gc, IncludeInferiors);
	XSetFunction(display, class.gc, GXinvert);
	XDrawRectangle(display, ROOT, class.gc, x1, y1, w, h);
	x1+=BORDERW-1;
	y1+=BORDERW-1;
	w-=(BORDERW-1)*2;
	h-=(BORDERW-1)*2;
	XDrawRectangle(display, ROOT, class.gc, x1, y1, w, h);

}
/****************************************************************************/




/*º Geometry helper functions ***********************************************/
static void take_a_trip (PAWindow pawindow)
{
	int x1=0, y1=0, x2=0, y2=0;
	int xoff=0, yoff=0;
	long evmask;
	unsigned int tmpmask;
	Window root=None;
	Window window=None;

	ASSERT(pawindow);

	/* Unmap the buttons... */
	button_unmap(class.button.close);
	button_unmap(class.button.iconz);
	button_unmap(class.button.maximize);

	/* Reset the saved geometry */
	pawindow->oldx=0; pawindow->oldy=0;
	pawindow->oldw=0; pawindow->oldh=0;

	/* Where the fuck is the fucking pointer???? */
	if (XQueryPointer(display, pawindow->frame, &root, &window,
					  &x1, &y1, &x2, &y2, &tmpmask)==False) BUG();

	if ((x1 < 0) || (y1 < 0)) BUG();

	/* Determine pointer offset */
	xoff=x1-pawindow->x;
	yoff=y1-pawindow->y;

	/* Determine corners for the outline */
	x1=pawindow->x;
	y1=pawindow->y;
	x2=pawindow->x+pawindow->w-1;
	y2=pawindow->y+pawindow->h-1;

	draw_outline(x1, y1, x2, y2);

	/* Special event mask for doing the move */
	evmask=ButtonPressMask|ButtonReleaseMask|PointerMotionMask;

	/* Capture both pointer and XServer . By the way, X data types are
	 * a huge pile of shit. Event mask is a long in X*Event functions, but 
	 * in XGrabPointer is an unsigned int */
	XGrabPointer(display, ROOT, True, (unsigned int)evmask,
				 GrabModeAsync, GrabModeAsync, None, CURSOR, CurrentTime);
	XGrabServer(display);

	while (1) {  /* Loops 'til button released */
		XEvent xevent;
		static Bool moved=False;

		XMaskEvent(display, evmask, &xevent);

		switch (xevent.type) {

			case MotionNotify: {  /* If pointer moved, move outline */
				XMotionEvent event=xevent.xmotion;

				moved=True;
				draw_outline(x1, y1, x2, y2);

				x1=event.x_root-xoff;
				y1=event.y_root-yoff;
				x2=x1+pawindow->w-1;
				y2=y1+pawindow->h-1;

				draw_outline(x1, y1, x2, y2);
				break;
			}

			case ButtonRelease: {  /* End the moving */
					
				XButtonPressedEvent *event=&(xevent.xbutton);
		   
				if (event->button!=Button1) break;

				draw_outline(x1, y1, x2, y2);

				/* Translate to client coordinates */
				if (pawindow->decoration) {
						x1+=BORDERW;
						y1+=BORDERW+TITLEBAR_H;
				}

				if (moved==True) pawindow_move(pawindow, x1, y1);

				button_map(class.button.close);
				button_map(class.button.iconz);
				button_map(class.button.maximize);

				XUngrabServer(display);
				XUngrabPointer(display, CurrentTime);
				return;
			}
        }
    }
}




/* Resize the pawindow */
static void stretch_and_shrink (PAWindow pawindow)
{
	/* Well, gravity is not needed at all */
	Cursor cursor=None;
	int x=0;
	int y=0;
	int x1=0, y1=0, x2=0, y2=0;  /* Outline corners */
	/* Variable aliases to improve readability... */
	int min_w=pawindow->min_w+BORDERW*2;
	int min_h=pawindow->min_h+BORDERW*2+TITLEBAR_H;
	int max_w=pawindow->max_w+BORDERW*2;
	int max_h=pawindow->max_h+BORDERW*2+TITLEBAR_H;
	unsigned int norm_w, norm_h;   /* Normalized width and height with size increments */
	Window root=None;
	Window window=None;
	unsigned int tmpmask;
	long evmask;

	ASSERT(pawindow);

	/* Avoid resizing if the pawindow is shaded */
	if (pawindow->shaded) return;

	/* Unmap the buttons... */
	button_unmap(class.button.close);
	button_unmap(class.button.iconz);
	button_unmap(class.button.maximize);

	/* Find the pointer */
	if (XQueryPointer(display, pawindow->frame, &root, &window,
					  &x1, &y1, &x, &y, &tmpmask)==False) BUG();

	if (x < 0) x=0;
	if (y < 0) y=0;

	/* Find the corner to warp pointer to */
	x=(x > (int)pawindow->w/2) ? pawindow->w-1 : 0;
	y=(y > (int)pawindow->h/2) ? pawindow->h-1 : 0;

	/* Put a pretty cursor depending on the corner ;) */
	cursor=(x==0) ? ((y==0) ? class.cursor.NW : class.cursor.SW):
					((y==0) ? class.cursor.NE : class.cursor.SE);

	/* Move the pointer to the nearest corner */
	XWarpPointer(display, pawindow->frame, pawindow->frame,
				 0, 0, pawindow->w, pawindow->h, x, y);

	/* Nice event mask to manage mouse temporarily */
	evmask=ButtonPressMask|ButtonReleaseMask|PointerMotionMask;

	/* Do the grabbings */
	XGrabPointer(display, ROOT, True, (unsigned int)evmask,
				 GrabModeAsync, GrabModeAsync, None,
				 cursor, CurrentTime);
	XGrabServer(display);

	/* Calculate corner coordinates */
	x1=pawindow->x;
	y1=pawindow->y;
	x2=pawindow->w+x1-1;
	y2=pawindow->h+y1-1;
	draw_outline(x1, y1, x2, y2);

	while (1) {  /* Loop 'til button released */
		XEvent xevent;
		XMaskEvent(display, evmask, &xevent);

		switch (xevent.type) {
			case MotionNotify: {
				XMotionEvent event=xevent.xmotion;

				draw_outline(x1, y1, x2, y2);

				x=event.x;
				y=event.y;

				/* Calculate outline on NorthWest corner */
				if (cursor==class.cursor.NW) {
					x1=x;
					y1=y;

					if (x2-x1 > max_w-1) x1=x2-max_w+1;
					if (x2-x1 < min_w-1) x1=x2-min_w+1;
					if (y2-y1 > max_h-1) y1=y2-max_h+1;
					if (y2-y1 < min_h-1) y1=y2-min_h+1;
				}
							
				/* Calculate outline on NorthEast corner */
				if (cursor==class.cursor.NE) {
					x2=x;
					y1=y;
									
					if (x2-x1 > max_w-1) x2=x1+max_w-1;
					if (x2-x1 < min_w-1) x2=x1+min_w-1;
					if (y2-y1 > max_h-1) y1=y2-max_h+1;
					if (y2-y1 < min_h-1) y1=y2-min_h+1;
				}
							
				/* Calculate outline on SouthWest corner */
				if (cursor==class.cursor.SW) {
					x1=x;
					y2=y;
									
					if (x2-x1 > max_w-1) x1=x2-max_w+1;
					if (x2-x1 < min_w-1) x1=x2-min_w+1;
					if (y2-y1 > max_h-1) y2=y1+max_h-1;
					if (y2-y1 < min_h-1) y2=y1+min_h-1;
				}
							
				/* Calculate outline on SouthEast corner */
				if (cursor==class.cursor.SE) {

					x2=x;
					y2=y;
								   
					if (x2-x1 > max_w-1) x2=x1+max_w-1;
					if (x2-x1 < min_w-1) x2=x1+min_w-1;
					if (y2-y1 > max_h-1) y2=y1+max_h-1;
					if (y2-y1 < min_h-1) y2=y1+min_h-1;
				}

				draw_outline(x1, y1, x2, y2);
				break;
        }

		case ButtonRelease: {
			XButtonPressedEvent *event=&(xevent.xbutton);
						
			if (event->button!=Button1) break;
			
			draw_outline(x1, y1, x2, y2);

			/* Translate to client coordinates */
			if (pawindow->decoration) {
					x1+=BORDERW;
					y1+=BORDERW+TITLEBAR_H;
					x2-=BORDERW;
					y2-=BORDERW;
			}

			/* Reset saved position before resizing */
			pawindow->oldx=0; pawindow->oldy=0;
			pawindow->oldw=0; pawindow->oldh=0;

			pawindow_move(pawindow, x1, y1);
			norm_w=x2-x1+1;
			norm_h=y2-y1+1;
			
			/* Honor size increments. Respecto */
			if (pawindow->w_inc) 
				while ((norm_w % pawindow->w_inc)!=0) {
					norm_w--;
				}
			if (pawindow->h_inc)	
				while ((norm_h % pawindow->h_inc)!=0) {
					norm_h--;
				}
			
			pawindow_resize(pawindow, norm_w, norm_h);
			pawindow_map(pawindow);
			XUngrabServer(display);
			XUngrabPointer(display, CurrentTime);
			return;
			}
		}
	}
}
/****************************************************************************/


/*º Other helpers ***********************************************************/
/* This one kills a client */
void kill_client (void *pw)
{
	PAWindow pawindow=(PAWindow)pw;
	Atom *proto=NULL;
	int n=0;

	if (!pw) return;

	/* See whether the client accepts WM protocols */        
	if (XGetWMProtocols(display, pawindow->window, &proto, &n)) {
		int i=0;
		for (i=0; i<n; i++) {
			if (proto[i]==atoms.wm_delete_window) {
				XEvent e;

				/* Build and send the WM_DELETE command to the client */
				SAY("WM_DELETE sent");
		 		e.type=ClientMessage;
				e.xclient.window=pawindow->window;
				e.xclient.message_type=atoms.wm_protocols;
				e.xclient.format=32;
				e.xclient.data.l[0]=atoms.wm_delete_window;
				e.xclient.data.l[1]=CurrentTime;

				if (!XSendEvent(display, pawindow->window, False,
						    NoEventMask, &e)) BUG();
				/* Shaded means we'll not get any damn UnmapNotify so...*/
				if (pawindow->shaded)
					pawindow_unmap(pawindow);
				break;
			}
			/* Atom not found, just kill */
			if (i==n) XKillClient(display, pawindow->window);
		}
	} else XKillClient(display, pawindow->window);  /* Fuck out */
}


void maximize_client (void *pw)
{

	PAWindow pawindow=(PAWindow)pw;
	char *tmpvalue;

	if (!pw) return;

	if (pawindow->shaded) 
		pawindow_unshade(pawindow);
	else if (!pawindow->oldw) {
		/* Keep previous geometry */
		pawindow->oldx=pawindow->x;
		pawindow->oldy=pawindow->y;
		pawindow->oldw=pawindow->w;
		pawindow->oldh=pawindow->h;

		/* Don't obscure the pabar when maximizing */
		tmpvalue=cfgpool_getstring(pawmpool, "pabar_autohide");
		if (strcmp(tmpvalue,"On")==0) 
				pawindow_resize(pawindow, DWIDTH-(BORDERW*2),
								  LOWBAR-(BORDERW*2)-TITLEBAR_H);
		else
				pawindow_resize(pawindow, DWIDTH-(BORDERW*2),
								  TOPBAR-(BORDERW*3)-TITLEBAR_H);
			/* Pawindow functions use client relative coordinates... */
			//if (pawindow->w!=pawindow->oldw || pawindow->h!=pawindow->oldh)
		pawindow_move(pawindow, BORDERW, BORDERW+TITLEBAR_H);

		free(tmpvalue);
	} else {        /* Restore previous geometry */
		pawindow->x=pawindow->oldx; pawindow->oldx=0;
		pawindow->y=pawindow->oldy; pawindow->oldy=0;
		pawindow->w=pawindow->oldw; pawindow->oldw=0;
		pawindow->h=pawindow->oldh; pawindow->oldh=0;

		pawindow_move(pawindow, pawindow->x+BORDERW, pawindow->y+BORDERW+TITLEBAR_H);
		pawindow_resize(pawindow, pawindow->w-(BORDERW*2),
					  pawindow->h-(BORDERW*2)-TITLEBAR_H);
	}

}



void iconify_client (void *pw)
{
	PAWindow pawindow=(PAWindow)pw;

	if (!pw) return;

	pawindow_unmap(pawindow);
}


/* Tab window helpers */
static void tab_window_setup(int num_windows)
{
	int num_rows=1;
	
	if (num_windows >6)
		num_rows=((num_windows-1)/6)+1;

	/* Resize tab windows */
	XResizeWindow(display, class.tab_window.tab, TAB_W, TAB_H);
	XMoveWindow(display, class.tab_window.tab, (DWIDTH-TAB_W)/2, (DHEIGHT-TAB_H)/2);
	XResizeWindow(display, class.tab_window.icons, TAB_ICONW_W, (num_rows*(48+TAB_ICON_SPACING))+(TAB_FRAME_PADDING*2));
	XMoveWindow(display, class.tab_window.text, TAB_MARGIN, TAB_MARGIN+(num_rows*(48+TAB_ICON_SPACING))+(TAB_FRAME_PADDING*2));
	XResizeWindow(display, class.tab_window.text, TAB_TEXTW_W, pafont_height(class.font));
}


static PAWindow tab_window_draw_all(int num_windows, int num_rows)
{
	PAWindow first_pw=NULL;
	PAWindow current_pw=NULL;
	int title_len=0;
   	unsigned int iconx, icony, firsticonx; int pos;

	if (num_rows > 1) {
		pos=TAB_MAX_ICONS_PERROW;
	} else {
		pos=num_windows;
	}
	
	/* Obtain the first pawindow in the tab list */
	chain_home(class.fpawindows);
	
	if (!focuspawindow) {
		while ((first_pw=chain_next(class.fpawindows))!=NULL) {
			if (first_pw && (first_pw->state==PA_MAPPED || first_pw->state==PA_ICONIFIED)) 
				/* We've found a selectable pawindow */
				break;
		}
	} else {
		while (chain_next(class.fpawindows)!=focuspawindow);
		first_pw=focuspawindow;
	}

	current_pw=first_pw;
	
	/* Draw the icons in the pop-up window */
	firsticonx=iconx=(TAB_ICONW_W-(pos*48)-((pos-1)*TAB_ICON_SPACING))/2;
	icony=TAB_FRAME_PADDING;
	pos=1;
	do {
		if (current_pw && (current_pw->state==PA_MAPPED || current_pw->state==PA_ICONIFIED)) {
			if (current_pw->wmicon->pixmap!=None) {	
				float scaledh, scaledw; float wf;
				XMoveWindow(display, current_pw->wmicon_win, iconx, icony);
				XMapRaised(display, current_pw->wmicon_win);
                XShapeCombineMask(display, current_pw->wmicon_win, ShapeBounding,
		                     0, 0, current_pw->wmicon->mask, ShapeSet);
				
				/* TODO: Optimize. Move this numbers to pawindow_create */
				/* Get the new propotions for scaled icons */
				if (current_pw->wmicon->w>=48) {
					wf=(float)current_pw->wmicon->w;
					if (current_pw->wmicon->w > current_pw->wmicon->h) {
						scaledw=ICONW;
						scaledh=(current_pw->wmicon->h/wf)*ICONW;
					} else {
						scaledh=ICONH;
						scaledw=(wf/current_pw->wmicon->h)*ICONW;
					}
				} else {
					scaledw=ICONW;
					scaledh=ICONH;
				}
				
				XCopyArea(display, current_pw->wmicon->pixmap, current_pw->wmicon_win, class.gc, 0, 0, scaledw, scaledh, 0, 0);
				
				if ( (pos % 6)!=0) {
					iconx+=(48+TAB_ICON_SPACING);
				} else {
					iconx=firsticonx;
					icony+=48+TAB_ICON_SPACING;					
				}
				pos++;
			}
		}
		current_pw=chain_next(class.fpawindows);
	} while ( current_pw != first_pw || !current_pw); 
   
	XSetClipMask(display, class.gc, None);

	/* Advance forward one position */
	chain_home(class.fpawindows);
	if (num_windows > 1) {
		chain_next(class.fpawindows);
		current_pw=chain_next(class.fpawindows);
	}
	
	/* Draw the title of the window */
	title_len=pafont_trimtext(class.font, current_pw->title, TAB_TEXTW_W);
	pafont_setcolor(class.font, class.color.textshadow);
	pafont_draw(class.font, class.tab_window.text,
			    (TAB_TEXTW_W-pafont_width(class.font, current_pw->title, title_len))/2, 0,
			    current_pw->title, title_len);

	return current_pw;
}


void rotate_clients(void *data)
{
	long evmask=KeyPressMask|KeyReleaseMask;
	unsigned int num_windows=0;
	PAWindow current_pw=NULL;
	int pos;
	int num_rows=1;
	int firsticonx, lasticonx, iconx, icony;
	bool forward= (strcmp((char *)data, "forward")==0 ? 1 : 0);


	/* TODO: Support both rotate method: w/wout popup window*/

	/* Get the number of created pawindows */
	chain_home(class.pawindows);
	while ((current_pw=chain_next(class.pawindows))!=NULL) {
		if (current_pw && (current_pw->state==PA_MAPPED || current_pw->state==PA_ICONIFIED)) {
			num_windows++;
		}
	}

	if (num_windows==0) return;

	if (num_windows > TAB_MAX_ICONS_PERROW) {
		pos= TAB_MAX_ICONS_PERROW;
		num_rows=((num_windows-1)/TAB_MAX_ICONS_PERROW)+1;
	} else {
		pos=num_windows;
	}
	
	tab_window_setup(num_windows);
	
	XMapWindow(display, class.tab_window.icons);
	XMapWindow(display, class.tab_window.text);
	XMapRaised(display, class.tab_window.tab);

	//XGrabServer(display);
    XGrabPointer(display, ROOT, True, 0, GrabModeAsync, GrabModeAsync, None,
                 None, CurrentTime);
	current_pw=tab_window_draw_all(num_windows, num_rows);

  	KeySym testkeysym; KeyCode testkeycode;
    testkeysym=XStringToKeysym("Tab");
    testkeycode=XKeysymToKeycode(display, testkeysym);
    XGrabKey(display, testkeycode, Mod1Mask, ROOT, False, GrabModeAsync, GrabModeAsync);
    XGrabKey(display, testkeycode, Mod1Mask|ShiftMask, ROOT, False, GrabModeAsync, GrabModeAsync);
    
	firsticonx=((TAB_ICONW_W-(pos*48)-((pos-1)*TAB_ICON_SPACING))/2)-TAB_FRAME_PADDING;
	if (num_rows==1) {
		lasticonx=firsticonx+((48+TAB_ICON_SPACING)*(num_windows-1));
	} else {
		lasticonx=firsticonx+((48+TAB_ICON_SPACING)*((num_windows % TAB_MAX_ICONS_PERROW)-1));
	}
		
	if (num_windows > 1) {
		if (forward) {
			/* Start at second icon */
			iconx=firsticonx+48+TAB_ICON_SPACING;
			icony=0;
		   	pos=2;
		} else {
			/* Start at last icon */
			iconx=lasticonx;
			icony=(48+TAB_ICON_SPACING)*(num_rows-1);
			pos=num_windows;
		}
	} else {
		iconx=firsticonx; icony=0;
		pos=1;
	}
	
	/* Draw a box in the next window, or in the first if there is only one */
	XSetFunction(display, class.gc, GXxor);
	XDrawRectangle(display, class.tab_window.icons, class.gc, iconx, icony, 53 ,53) ;
	XDrawRectangle(display, class.tab_window.icons, class.gc, iconx+1, icony+1, 51 ,51) ;
	XSetFunction(display, class.gc, GXcopy);
	
    
	while (1) {  
        XEvent xevent;
		
        XMaskEvent(display, evmask, &xevent);

        switch (xevent.type) {
        case KeyPress: {
            XKeyPressedEvent event=xevent.xkey;
			int title_len=0;

			/* Avoid unnecesary work */
			if (num_windows==1) break;
			
			SAY("loop: Got Tab press...");
			if (!keyboard_compare_key("Tab", Mod1Mask, &event)) {
				SAY("Key is not tab");
				break;
			} else if (keyboard_compare_key("Tab", Mod1Mask|ShiftMask, &event)) {
				forward=0;
			} else if (keyboard_compare_key("Tab", Mod1Mask, &event)) {
				forward=1;
			} 	
			
			do {
				current_pw=chain_next(class.fpawindows);
			} while (!current_pw || (current_pw->state!=PA_MAPPED && current_pw->state!=PA_ICONIFIED));
			
			XClearWindow(display, class.tab_window.text);
			pafont_setcolor(class.font, class.color.textshadow);
			title_len=pafont_trimtext(class.font, current_pw->title, TAB_TEXTW_W);
			pafont_draw(class.font, class.tab_window.text,
						(TAB_TEXTW_W-pafont_width(class.font, current_pw->title, title_len))/2, 0,
						current_pw->title, title_len);
			
			/* Undraw the previous box */
			XSetFunction(display, class.gc, GXxor);
			XDrawRectangle(display, class.tab_window.icons, class.gc, iconx, icony, 53 ,53) ;
			XDrawRectangle(display, class.tab_window.icons, class.gc, iconx+1, icony+1 , 51 ,51) ;
			
			/* Move box to next position */
			if (forward) {
				if (pos==num_windows) {  /* Wrap around */
					iconx=firsticonx;
					icony=0;
					pos=1;
				} else {
					if ( (pos==1) || (pos % 6)!=0) {
						iconx+=48+TAB_ICON_SPACING;
					} else {
						iconx=firsticonx;
						icony+=48+TAB_ICON_SPACING;					
					}
						
					pos++;
				}
			} else { /* Backward */	
				if (pos==1) {			/* Wrap around */
					iconx=lasticonx;  
					icony=(48+TAB_ICON_SPACING)*(num_rows-1);
					pos=num_windows;
				} else {
					if ( ((pos-1) % 6)!=0) {
						iconx-=48+TAB_ICON_SPACING;
					} else {
						iconx+=(48+TAB_ICON_SPACING)*(TAB_MAX_ICONS_PERROW-1); 
						icony-=48+TAB_ICON_SPACING;					
					}
					pos--;
				}
			}
			
			/* Draw the next box */
			XDrawRectangle(display, class.tab_window.icons, class.gc, iconx, icony , 53 ,53) ;
			XDrawRectangle(display, class.tab_window.icons, class.gc, iconx+1, icony+1 , 51 ,51) ;
			XSetFunction(display, class.gc, GXcopy);
			
			break;
	    }
		case KeyRelease: {
            XKeyReleasedEvent event=xevent.xkey;
			SAY("loop: Got release... %d", event.keycode);
			if (keyboard_compare_key("Alt_L", AnyModifier, &event)) {
				SAY("loop: Got alt release");
				XUnmapWindow(display, class.tab_window.tab);
				/* Map only if needed */
				if (current_pw!=focuspawindow) 
				   	pawindow_map(current_pw);
				/* Ungrab alt key */
				XUngrabKey(display, testkeycode, Mod1Mask, ROOT);
				XUngrabPointer(display, CurrentTime);
				//XUngrabServer(display);
				return;
			}
	    }
		}	
   }

}


static void restore_client (void *pw)
{
	PAWindow pawindow=(PAWindow)pw;

	ASSERT(pw);

	if ((pawindow->state == PA_MAPPED) && pawindow->active) 
		pawindow_unmap(pawindow);
	else 
		pawindow_map(pawindow);
    
}


/* This function sets up a new title onto the title bar, in UTF-8 */
static void set_title (PAWindow pawindow, char *text, size_t nbytes)
{
	ASSERT(pawindow);
	ASSERT(text);
	    
	if (pawindow->title) free(pawindow->title);
	pawindow->title=NULL;

	/* Copy the string */
	pawindow->title=malloc(nbytes+1);
	if (!(pawindow->title)) die("Not enough memory");  /* Huy que mal rollito... */

	strncpy(pawindow->title, text, nbytes);
	pawindow->title[nbytes]='\0';

	draw_titlebar(pawindow);

}


/* Give the keyboard focus to the PAwindow passed as parameter */
static void change_focus(PAWindow pawindow) 
{
	PAWindow pw;

	ASSERT(pawindow);

	/* Don't set the keyboard focus in utility windows */
	if (pawindow->utility) return;
	
	/* The pawindow already has the focus, give it again */
	if (focuspawindow && focuspawindow==pawindow) {
		if (pawindow->shaded)
			XSetInputFocus(display, pawindow->frame, RevertToPointerRoot, CurrentTime);
		else
			XSetInputFocus(display, pawindow->window, RevertToParent, CurrentTime);
		return;
	}

	if (focuspawindow) {
		focuspawindow->active=False;
		if (focuspawindow->icon) palabel_unsetfocus(focuspawindow->icon);
		/* A BadWindow could happen here (could==always in X Window notation),
		 * when the unmapped window has been already destroyed by the client. 
		 * X kicking asses ;) in ye good olde tradition */
		draw_titlebar(focuspawindow);
	}

	/* Remove previous references... */
	chain_home(class.fpawindows);
	while( (pw=chain_next(class.fpawindows))!=NULL ) {
		if (pw==pawindow || pw->state==PA_WITHDRAWN) 
	   		chain_del(class.fpawindows);
	}

	focuspawindow=pawindow;
	SAY("Pawindow %#lx pushed",pawindow->window);
	chain_push(class.fpawindows, pawindow);

	if (pawindow->shaded) 
		XSetInputFocus(display, pawindow->frame, RevertToPointerRoot, CurrentTime);
	else 
		XSetInputFocus(display, pawindow->window, RevertToParent, CurrentTime);

	palabel_setfocus(pawindow->icon);

	x_set_property(ROOT, atoms.net_active_window, XA_WINDOW, (unsigned char *)&(pawindow->window), 1);

}


/* Configures the shape of a pawindow, if needed */
static void set_shape(PAWindow pawindow)
{
	XRectangle *shp_rectangles;    /* Shape information */
	int num,order;
	int major,minor;

	ASSERT(pawindow);

	XShapeQueryVersion(display, &major, &minor);
	SAY("Configuring shape... (XShape version %d.%d)",major,minor);

	/* First we get the shape of the client window, and after use it to set
	   the shape of the pawindow canvas and frame. We don't do a ShapeClip
	   because X window borders are not being used by pawm */
	SAY("Setting bounding shape in pawindow");
	shp_rectangles=XShapeGetRectangles(display, pawindow->window, 
					   ShapeBounding, &num, &order);
	if (shp_rectangles==NULL) return;
	SAY("set_shape: shaping canvas...");
	XShapeCombineRectangles(display, pawindow->canvas, ShapeBounding,
				0, 0, shp_rectangles, num, ShapeSet, order);
	SAY("set_shape: shaping frame...");
	XShapeCombineRectangles(display, pawindow->frame, ShapeBounding,
			    0, 0, shp_rectangles, num, ShapeSet, order);
	SAY("set_shape: shaping done");
	XFree(shp_rectangles);
	SAY("set_shape: exiting...");

}

static void set_client_list()
{
	/* Add windows to _NET_CLIENT_LIST */
	Window *winlist;
	PAWindow pw;
	int i;

	winlist=malloc(chain_size(class.pawindows)*sizeof(Window));
	if (!winlist) die("Not enough memory");

	if (chain_size(class.pawindows)==0) {
		x_set_property(ROOT, atoms.net_client_list, XA_WINDOW, NULL, 0);
	}

	chain_home(class.pawindows);
	for (i=0; i<chain_size(class.pawindows); i++) {
		pw=chain_next(class.pawindows);
		winlist[i]=pw->window;
	}
	x_set_property(ROOT, atoms.net_client_list, XA_WINDOW, 
		      (unsigned char *)winlist, chain_size(class.pawindows));

	free(winlist);
}


static void set_fullscreen(PAWindow pawindow, bool state)
{

	if (state) {
		pawindow->oldfx=pawindow->x;
		pawindow->oldfy=pawindow->y;
		pawindow->oldfw=pawindow->w;
		pawindow->oldfh=pawindow->h;
		pabar_unmap();
		pawindow_move(pawindow, 0, 0);
		pawindow_resize(pawindow, DWIDTH, DHEIGHT);
	} else {
		/* Restore previous geometry */
		if (pawindow->oldfx) {
			pawindow_resize(pawindow, pawindow->oldfw-(2*BORDERW), pawindow->oldfh-(2*BORDERW+TITLEBAR_H));
			pawindow_move(pawindow, pawindow->oldfx+BORDERW, pawindow->oldfy+(BORDERW+TITLEBAR_H));
			pawindow->oldfx=0;
			pawindow->oldfy=0;
			pawindow->oldfw=0;
			pawindow->oldfh=0;
		}
	}
   
}


/* Send a ConfigureNotify to the client window */
static void notify_client(PAWindow pawindow)
{
	XEvent e;

	/* Build and send the notice for our client - ICCCM */
	/* Watch the fuck out! If the pawindow is shaded, only the titlebar
	   is visible, so we DON'T want to send a ConfigureNotify to the client 
	   or strange things are going to happen */
	if (!pawindow->shaded) {
		e.type=ConfigureNotify;
		e.xconfigure.event=pawindow->window;
		e.xconfigure.window=pawindow->window;
		if (pawindow->decoration) {
			e.xconfigure.x=pawindow->x+BORDERW;
			e.xconfigure.y=pawindow->y+TITLEBAR_H+BORDERW;
			e.xconfigure.width = pawindow->w-(BORDERW*2);
			e.xconfigure.height= pawindow->h-TITLEBAR_H-(BORDERW*2);
		} else {
			e.xconfigure.x=pawindow->x;
			e.xconfigure.y=pawindow->y;
			e.xconfigure.width = pawindow->w;
			e.xconfigure.height= pawindow->h;
		}
		e.xconfigure.border_width=0;
		e.xconfigure.above=None;
		e.xconfigure.override_redirect=False;

		SAY("Sending ConfigureNotify");
		if (!XSendEvent(display, pawindow->window, False,
				StructureNotifyMask, &e)) BUG();
	}
}


void set_scale_down_matrix(double factor)
{
	/* Scaling matrix for icons */
	class.scale_matrix.matrix[0][0]=XDoubleToFixed(1);
	class.scale_matrix.matrix[0][1]=XDoubleToFixed(0);
	class.scale_matrix.matrix[0][2]=XDoubleToFixed(0);
	class.scale_matrix.matrix[1][0]=XDoubleToFixed(0);
	class.scale_matrix.matrix[1][1]=XDoubleToFixed(1);
	class.scale_matrix.matrix[1][2]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][0]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][1]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][2]=XDoubleToFixed(factor);
}

void set_scale_up_matrix(double factor)
{
	/* Scaling matrix for icons */
	class.scale_matrix.matrix[0][0]=XDoubleToFixed(factor);
	class.scale_matrix.matrix[0][1]=XDoubleToFixed(0);
	class.scale_matrix.matrix[0][2]=XDoubleToFixed(0);
	class.scale_matrix.matrix[1][0]=XDoubleToFixed(0);
	class.scale_matrix.matrix[1][1]=XDoubleToFixed(factor);
	class.scale_matrix.matrix[1][2]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][0]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][1]=XDoubleToFixed(0);
	class.scale_matrix.matrix[2][2]=XDoubleToFixed(1);
}


/****************************************************************************/




/****************************************************************************/
/*                                                                          */
/*                        Public (member) functions                         */
/*                                                                          */
/****************************************************************************/


/*
        This function initializes the class, not the object. DO NOT EVER
instantiate a PAWindow object without initializing its class...
*/
void pawindow_init (void)
{
	char *tmpvalue;

	if (reference_count++) return;

	/* Grab keybindings for the root window */
	keyboard_grabkeys(ROOT);

	/* Get font, since its size will determine the minimum height */
	class.font=pafont_create(display, "pawindow");

	/* Get the centered position */
	class.texty=(TITLEBAR_H-pafont_height(class.font))/2;

	class.gc=XCreateGC(display, ROOT, 0, NULL);
	if (class.gc==None) BUG();

	XSetForeground(display, class.gc, BLACK);
	XSetGraphicsExposures(display, class.gc, False);

	class.color.passivebar=x_color("titlebar_inactive");
	class.color.activebar=x_color("titlebar_active");
	class.color.passivetext=x_colorstr("text_inactive");
	class.color.activetext=x_colorstr("text_active");
	class.color.textshadow=x_colorstr("text_shadow");
	class.color.framelight=x_color("frame_light");
	class.color.framewhite=x_color("frame_white");
	class.color.framebase=x_color("frame_base_color");
	class.color.framedark=x_color("frame_dark");
	class.color.framedarker=x_color("frame_darker");


	/* Build the pointers */
	class.cursor.NW=XCreateFontCursor(display, NWCURSOR);
	if (class.cursor.NW==None) BUG();
	class.cursor.NE=XCreateFontCursor(display, NECURSOR);
	if (class.cursor.NE==None) BUG();
	class.cursor.SW=XCreateFontCursor(display, SWCURSOR);
	if (class.cursor.SW==None) BUG();
	class.cursor.SE=XCreateFontCursor(display, SECURSOR);
	if (class.cursor.SE==None) BUG();
	class.cursor.resize=XCreateFontCursor(display, RESIZECURSOR);
	if (class.cursor.resize==None) BUG();

	/* Build the button faces, masks aren't used for buttons, by now ;)) */
	tmpvalue=cfgpool_getstring(pawmpool,"close_normal");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "close_normal");
	class.face.close_normal=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.close_normal==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"close_active");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "close_active");
	class.face.close_active=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.close_active==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"close_pressed");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "close_pressed");
	class.face.close_pressed=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.close_pressed==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"min_normal");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "min_normal");
	class.face.iconz_normal=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.iconz_normal==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"min_active");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "min_active");
	class.face.iconz_active=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.iconz_active==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"min_pressed");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "min_pressed");
	class.face.iconz_pressed=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.iconz_pressed==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"max_normal");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "max_normal");
	class.face.maximize_normal=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.maximize_normal==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"max_active");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "max_active");
	class.face.maximize_active=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.maximize_active==NULL) BUG();

	tmpvalue=cfgpool_getstring(pawmpool,"max_pressed");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "max_pressed");
	class.face.maximize_pressed=x_build_pixmap(tmpvalue);
	free(tmpvalue);
	if (class.face.maximize_pressed==NULL) BUG();

	/* Check shadow */ 
	tmpvalue=cfgpool_getstring(pawmpool,"window_shadow");
	if (tmpvalue==NULL) die(NOTFOUNDMSG, "window_shadow");
	if (strcmp(tmpvalue,"On")==0)
		class.shadow_status=True;
	else
		class.shadow_status=False;

	/* No window has the focus ;) */
	focuspawindow=NULL;

	class.timestamp=0;
	class.clickx=-1;
	class.clicky=-1;	

	/* Initialize other needed classes */
	button_init();

	class.button.iconz=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ, class.face.iconz_normal->pixmap,
									class.face.iconz_active->pixmap, class.face.iconz_pressed->pixmap,
									iconify_client, NULL);


	class.button.close=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ, class.face.close_normal->pixmap,
									 class.face.close_active->pixmap, class.face.close_pressed->pixmap,
									 kill_client, NULL);

	class.button.maximize=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ, class.face.maximize_normal->pixmap,
										class.face.maximize_active->pixmap, class.face.maximize_pressed->pixmap,
										NULL, NULL);

	class.pawindows=chain_create();
	if (!class.pawindows) BUG();

	class.fpawindows=chain_create();
	if (!class.fpawindows) BUG();

	/* Create empty tab-windows (to have a place to reparent wmicons) */
	class.tab_window.tab=XCreateSimpleWindow(display, ROOT, 0, 0, 1, 1, 1, class.color.framedark, class.color.framebase);
	class.tab_window.icons=XCreateSimpleWindow(display, class.tab_window.tab, TAB_MARGIN, TAB_MARGIN, 1, 1, 0, BLACK, class.color.framebase);
	class.tab_window.text=XCreateSimpleWindow(display, class.tab_window.tab, 0, 0, 1, 1, 0, BLACK, class.color.framebase);
}




/* This one destroys the class */
void pawindow_done (void)
{ 
	PAWindow pawindow;	

	if (--reference_count) return;

	XFreeCursor(display, class.cursor.NW);
	XFreeCursor(display, class.cursor.NE);
	XFreeCursor(display, class.cursor.SW);
	XFreeCursor(display, class.cursor.SE);
	XFreeCursor(display, class.cursor.resize);

	XFreeGC(display, class.gc);
	x_destroy_pixmap(class.face.close_normal);
	x_destroy_pixmap(class.face.close_active);
	x_destroy_pixmap(class.face.close_pressed);
	x_destroy_pixmap(class.face.iconz_normal);
	x_destroy_pixmap(class.face.iconz_active);
	x_destroy_pixmap(class.face.iconz_pressed);
	x_destroy_pixmap(class.face.maximize_normal);
	x_destroy_pixmap(class.face.maximize_active);
	x_destroy_pixmap(class.face.maximize_pressed);

	while ( (pawindow=chain_pop(class.fpawindows))!=NULL ); 

	chain_seek(class.pawindows, chain_size(class.pawindows));
	while ( (pawindow=chain_del(class.pawindows))!=NULL ) {
		pawindow_destroy(pawindow);
	    XSync(display, False);
	}

	chain_delete(class.fpawindows);
	chain_delete(class.pawindows);

	pafont_destroy(class.font);

	button_destroy(class.button.close);
	button_destroy(class.button.iconz);
	button_destroy(class.button.maximize);

}




/* Makes a new pawindow from the window, but DOESN'T map it. */
PAWindow pawindow_create (Window window) 
{
	PAWindow pawindow;
	Window parent=None;
	PAWindow pa_transientparent=NULL;
	XWindowAttributes attribs;
	XSizeHints *szHints=NULL;
	long mask=0;
	Bool bound=False,clip=False;
	int foo;
	unsigned long length;
	unsigned char *text;
	CARD32 **icons;

		
	ASSERT(window);

	/** Basic initialization **/
	pawindow=malloc(PAWindowBytes);
	if (!pawindow) die("Not enough memory");

	/* Register newly created window */
	XAddToSaveSet(display, window);
	chain_seek(class.pawindows, chain_size(class.pawindows));
	chain_put(class.pawindows, pawindow);

	/* Set our pawindow keybindings */
	keyboard_grabkeys(window);

	pawindow->window=window;
	pawindow->title=NULL;
	pawindow->modalchild=NULL;

	pawindow->transient=false;
	pawindow->active=false;
	pawindow->state=PA_WITHDRAWN;      /* ICCCM */
	pawindow->shaded=false;
	pawindow->got_netwmname=false;         /* EWMH */
	pawindow->got_netwmiconname=false;     /* EWMH */

	/* Does it need decoration? */
	pawindow->decoration=gtkHints_needDecoration(pawindow->window);

	/* Does it have a shape? */
	XShapeQueryExtents(display,pawindow->window, &bound,&foo,&foo,(unsigned int *)&foo,
		   				(unsigned int *)&foo, &clip,&foo,&foo, (unsigned int *)&foo, (unsigned int *)&foo);
	pawindow->shaped=bound;

	szHints=XAllocSizeHints();  /* Absolutely stupid... */
	if (!szHints) BUG();
		
	XSetWindowBorderWidth(display, window, 0);  /* NO border */

	pawindow->oldx=0; pawindow->oldy=0;
	pawindow->oldw=0; pawindow->oldh=0;
	pawindow->oldfx=0; pawindow->oldfy=0;
	pawindow->oldfw=0; pawindow->oldfh=0;
	pawindow->w_inc=0; pawindow->h_inc=0;

	pawindow_get_type(pawindow);
	
	length=0;
	icons=(CARD32 **)x_get_property(pawindow->window, atoms.net_wm_icon, XA_CARDINAL, &length);
	/* Use _NET_WM_ICON or a default icon if it isn't found */
	if (length >0) {
		pawindow->wmicon=x_build_pixmap_from_argb(icons, length);
		pawindow->wmpicture=XRenderCreatePicture(display, pawindow->wmicon->pixmap, RENDERFORMAT, 0, 0);
		/* TODO: check behaviour with icons less than 48 */
		if (pawindow->wmicon->w >= ICONW || pawindow->wmicon->h >= ICONH ) {
			/* Use the larger coordinate */
			if (pawindow->wmicon->w >= pawindow->wmicon->h)
				set_scale_down_matrix(48.0/pawindow->wmicon->w);
			else
				set_scale_down_matrix(48.0/pawindow->wmicon->h);
		} else if (pawindow->wmicon->w < ICONW ) {
			if (pawindow->wmicon->w >= pawindow->wmicon->h)
				//set_scale_up_matrix(pawindow->wmicon->w/48.0);
				set_scale_up_matrix(1);
			else
				set_scale_up_matrix(1);
		}

		XRenderSetPictureTransform(display, pawindow->wmpicture, &class.scale_matrix);
		XRenderSetPictureFilter(display, pawindow->wmpicture, FilterBilinear, 0, 0);
		XRenderComposite(display, PictOpOver, pawindow->wmpicture, None, pawindow->wmpicture,
						 0, 0, ICONW, ICONH, 0, 0, pawindow->wmicon->w, pawindow->wmicon->h);
	} else {
		pawindow->wmicon=x_build_pixmap(DEFAULT_ICON);
	}

	pawindow->wmicon_win=XCreateSimpleWindow(display, class.tab_window.icons, 0, 0, 48, 48,
												   0, BLACK, class.color.framebase);

	
	/* See if the window is transient or not */
	XGetTransientForHint(display, pawindow->window, &parent);
	pawindow->transient=(parent==None) ? false : true;

	if (pawindow->transient) {
			chain_home(class.pawindows);
			while ((pa_transientparent=chain_next(class.pawindows)) != NULL) {
				if (pa_transientparent->window==parent) { SAY ("FOUND");break;}
			}
	} 

	if (pawindow->dialog && pawindow->modal && pa_transientparent) {
		pa_transientparent->modalchild=pawindow;
	}

	/* Calculate sane values for geometry */
	if (XGetWindowAttributes(display, window, &attribs)) {
		if (!attribs.x && !attribs.y) {
			/* If not default values are given, make the pawindows 'tile' to
			   make them more visible and not to be overlapped each other */
			pawindow->x=(chain_size(class.pawindows)*10)-10;
			pawindow->y=(chain_size(class.pawindows)*10)-10;
		} else {
			if (pawindow->decoration) {
				pawindow->x=attribs.x-BORDERW;
				pawindow->y=attribs.y-BORDERW-TITLEBAR_H;
			} else {
				pawindow->x=attribs.x;
				pawindow->y=attribs.y;
			}
		}
		/* Center transient pawindows relative to parent and dialogs
		 * relative to screen */
		if (pawindow->transient && pa_transientparent) {
			if (attribs.width < pa_transientparent->w) {
				pawindow->x=((pa_transientparent->w - attribs.width -(BORDERW*2))/2) + pa_transientparent->x;
				pawindow->y=((pa_transientparent->h - attribs.height -(TITLEBAR_H+(BORDERW*2)))/2) + pa_transientparent->y;
			} else {
				pawindow->x= pa_transientparent->x - ((attribs.width-pa_transientparent->w)/2);
				pawindow->y=((pa_transientparent->h - attribs.height -(TITLEBAR_H+(BORDERW*2)))/2) + pa_transientparent->y;
			}
		} else if (pawindow->dialog) {
			pawindow->x=(DWIDTH-attribs.width-(BORDERW*2))/2;
			pawindow->y=(DHEIGHT-attribs.height-(TITLEBAR_H+(BORDERW*2)))/2;
		}
		pawindow->w=attribs.width;
		pawindow->h=attribs.height;
	} else {
	/* Set default values */    
		pawindow->x=(chain_size(class.pawindows)*10)-10;
		pawindow->y=(chain_size(class.pawindows)*10)-10;
		pawindow->w=FRAME_MIN_W;
		pawindow->h=FRAME_MIN_H;
	}

	pawindow->min_w=FRAME_MIN_W;
	pawindow->min_h=FRAME_MIN_H;
	pawindow->max_w=FRAME_MAX_W;
	pawindow->max_h=FRAME_MAX_H;


	/* Translate to the frame world */
	if (pawindow->decoration) {
		pawindow->w+=BORDERW*2;
		pawindow->h+=BORDERW*2+TITLEBAR_H;
	}

	 /* Make titlebar visible at creation time */
	if (pawindow->x < 0) pawindow->x=0;
	if (pawindow->y < 0) pawindow->y=0;

	/* Make width & height not bigger than display size at creation time */
	if (pawindow->w > DWIDTH)  pawindow->w=DWIDTH;
	if (pawindow->h > DHEIGHT) pawindow->h=DHEIGHT;

	/* Another check */
	if ( pawindow->w > (DWIDTH-pawindow->x) )  pawindow->x=0;
	if ( pawindow->h > (DHEIGHT-pawindow->y) ) pawindow->y=0;

		
	/* At this point, X and Y for the frame are fully calculated */


	/*
		NOTE: the PA WM calculates *ALL* geometries based upon
	the frame geometry, including the client geometry, etc...
	*/

	if (!XGetWMNormalHints(display, window, szHints, &mask)) szHints->flags=0;

	/* See whether we have size limits (client relative) */
	if (szHints->flags & PMinSize) {
		pawindow->min_w=szHints->min_width;
		pawindow->min_h=szHints->min_height;

		if (pawindow->decoration) {
			if (pawindow->min_w < FRAME_MIN_W) pawindow->min_w=FRAME_MIN_W;
			if (pawindow->min_h < FRAME_MIN_H) pawindow->min_h=FRAME_MIN_H;
			if (pawindow->min_w > FRAME_MAX_W) pawindow->min_w=FRAME_MAX_W;
			if (pawindow->min_h > FRAME_MAX_H) pawindow->min_h=FRAME_MAX_H;
		}
	}

	if (szHints->flags & PMaxSize) {
		pawindow->max_w=szHints->max_width;
		pawindow->max_h=szHints->max_height;

		if (pawindow->max_w < FRAME_MIN_W) pawindow->max_w=FRAME_MIN_W;
		if (pawindow->max_h < FRAME_MIN_H) pawindow->max_h=FRAME_MIN_H;
		if (pawindow->max_w > FRAME_MAX_W) pawindow->max_w=FRAME_MAX_W;
		if (pawindow->max_h > FRAME_MAX_H) pawindow->max_h=FRAME_MAX_H;
	}

	if (szHints->flags & PResizeInc) {
		pawindow->w_inc=szHints->width_inc;
		pawindow->h_inc=szHints->height_inc;
	}

	XFree(szHints);

	/* Sanitize values */
	if (pawindow->max_w < pawindow->min_w) pawindow->max_w=pawindow->min_w;
	if (pawindow->max_h < pawindow->min_h) pawindow->max_h=pawindow->min_h;

	/** Create the frame and set up decorations **/
	/* The shadow is, obviously, under the frame */

	/* Shadow window is created only when there is decoration */
	if (pawindow->decoration && class.shadow_status) {
		pawindow->shadow=XCreateSimpleWindow(display, ROOT, SHADOW_X, SHADOW_Y,
											 SHADOW_W, SHADOW_H, 0, 0, BLACK);
	if (pawindow->shadow==None) BUG();
		XStoreName(display, pawindow->shadow, "shadow");
		XDefineCursor(display, pawindow->shadow, CURSOR);
	}


	/* The frame stores all decorations */
	pawindow->frame=XCreateSimpleWindow(display, ROOT,
										pawindow->x, pawindow->y,
										pawindow->w, pawindow->h,
										0, BLACK, BLACK);
	if (pawindow->frame==None) BUG();
	XStoreName(display, pawindow->frame, "frame");
	XDefineCursor(display, pawindow->frame, class.cursor.resize);


	/* The canvas just serve as a nest for the client */
	if (pawindow->decoration) {
		pawindow->canvas=XCreateSimpleWindow(display, pawindow->frame, CANVAS_X, CANVAS_Y,
											 CANVAS_W, CANVAS_H, 0, BLACK, BLACK);
	} else {
		pawindow->canvas=XCreateSimpleWindow(display, pawindow->frame, 0, 0,
											 pawindow->w, pawindow->h, 0, BLACK, BLACK);
	}
	if (pawindow->canvas==None) BUG();

	XStoreName(display, pawindow->canvas, "canvas");
	XDefineCursor(display, pawindow->canvas, CURSOR);

	if (pawindow->decoration) {
		/* The titlebar... */
		pawindow->titlebar=XCreateSimpleWindow(display, pawindow->frame, TITLEBAR_X, TITLEBAR_Y,
											   TITLEBAR_W, TITLEBAR_H, 0, BLACK, class.color.passivebar);
		if (pawindow->titlebar==None) BUG();
		XStoreName(display, pawindow->titlebar, "titlebar");
		XDefineCursor(display, pawindow->titlebar, CURSOR);
	}

	/* Select events and register the Sloots */
	XSelectInput(display, pawindow->canvas, CANVAS_EVMASK);
	/* I love to use libraries with NO fucking documentation (ok, you
	 * can found a scarce 7 page pdf and the man pages with the prototypes
	 * for XShape, but that's all), another extensions (Xft comes to mind) 
	 * may be in the same state... Note: Yes, they are. XRender another nice example...*/
	XShapeSelectInput(display, pawindow->canvas, ShapeNotifyMask);
	if (!pawindow->decoration && pawindow->shaped) 
		set_shape(pawindow);    
	x_addsloot(pawindow->canvas, canvas_handle, pawindow);

	/* This handlers are only for decorated windows */
	if (pawindow->decoration) {
		XSelectInput(display, pawindow->titlebar, TITLEBAR_EVMASK);
		x_addsloot(pawindow->titlebar, titlebar_handle, pawindow);
	}

	/* This handler cannot be ignored for non-decorated windows, we
	   need it to implement click-to-focus policy */
	XSelectInput(display, pawindow->frame, FRAME_EVMASK);
	x_addsloot(pawindow->frame, frame_handle, pawindow);

	/*  This is somewhat special, since we shouldn't install a
	handler for our client window ;)) and we just want one event...
	So, this event (PropertyNotify), will be handled by the 'canvas'
	*/
	XSelectInput(display, pawindow->window, PropertyChangeMask);
	x_addsloot(pawindow->window, canvas_handle, pawindow);

	/* Reparent the client. Some stupid clients insist on being
	mapped out of screen and be here even if reparented. Well, we
	insist on moving them to the origin }:)))).  
	*/
	XReparentWindow(display, window, pawindow->canvas, 0, 0);

	pawindow->icon=palabel_create(200, restore_client, pawindow);

	/* Set first _NET_WM_NAME, or WM_NAME if the first is not available */
	SAY("Setting window name hint");
	text=x_get_property(pawindow->window, atoms.net_wm_name, atoms.utf8_string, &length);
	if (text) {
		set_title(pawindow, (char *)text, length);
	} else {	
		text=x_get_property(pawindow->window, atoms.wm_name, atoms.string, &length);
		if (text) set_title(pawindow, (char *)text, length);
	}
	if (text) XFree(text);

	/* Set first _NET_WM_ICON_NAME, or WM_ICON_NAME if the first is not available */
	SAY("Setting window icon hint");
	text=x_get_property(pawindow->window, atoms.net_wm_icon_name, atoms.utf8_string, &length);
	if (text) {
		palabel_settext(pawindow->icon, (char *)text, length);
	} else {	
		text=x_get_property(pawindow->window, atoms.wm_icon_name, atoms.string, &length);
		if (text) {
			palabel_settext(pawindow->icon, (char *)text, length);
		} else {
			/* Some libraries don't set the ICON_NAME. Qt is one fucking
			   nice example of how to ignore ICCCM and EWMH conventions... */
			palabel_settext(pawindow->icon, pawindow->title, (pawindow->title)?strlen(pawindow->title):0);
		}
	}	 
	if (text) XFree(text);

	pawindow_get_state(pawindow);

	/* Update client windows list */
	set_client_list();
			
	return pawindow;
}


/* Get the client window state, using the _NET_WM_STATE property atom. There
   is no drawing here, but in pawindow_map */
static void pawindow_get_state (PAWindow pawindow)
{
	CARD32 *atom_list;
	unsigned long length;
	int i;
	
	pawindow->fullscreen=false;
	pawindow->got_skiptaskbar=false;
	pawindow->modal=false;
	
	atom_list=(CARD32 *)x_get_property(pawindow->window, atoms.net_wm_state, XA_ATOM, &length);

	/* No atoms found, don't change anything */
	if (atom_list==NULL) return;
	
	for (i=0;i<length;i++) {
		SAY("Window asks for %s",XGetAtomName(display, atom_list[i]));
		if (atom_list[i]==atoms.net_wm_state_fullscreen) 
			pawindow->fullscreen=true;
		else if (atom_list[i]==atoms.net_wm_state_skip_taskbar)
			pawindow->got_skiptaskbar=true;
		else if (atom_list[i]==atoms.net_wm_state_modal)
			pawindow->modal=true;
	}
}


/* Get the client window type, using the _NET_WM_WINDOW_TYPE property atom.
   This is used later by pawm to know how to manage this window */
static void pawindow_get_type (PAWindow pawindow)
{
	CARD32 *atom_list;
	unsigned long length;
	int i;
	
	pawindow->dialog=false;
	pawindow->utility=false;

	atom_list=(CARD32 *)x_get_property(pawindow->window, atoms.net_wm_window_type, XA_ATOM, &length);
	for (i=0; i<length; i++) {
		SAY("Window wannabe a %s", XGetAtomName(display, atom_list[i]));
		if (atom_list[i]==atoms.net_wm_window_type_dialog) 
			pawindow->dialog=true;
		else if (atom_list[i]==atoms.net_wm_window_type_splash)
			pawindow->decoration=false;
		else if (atom_list[i]==atoms.net_wm_window_type_utility)
			pawindow->utility=true;;
	}
}



/* This entirely destroys the specified pawindow */
void pawindow_destroy (PAWindow pawindow)
{
	Window root;Window parent;
	Window *tree;
	unsigned int nchildren;
	PAWindow pw;

	ASSERT(pawindow);

	SAY("Someone wants to destroy %#lx", pawindow->window);
	if (pawindow->title) free(pawindow->title);

	if (pawindow->decoration) {
		x_delsloot(pawindow->frame, frame_handle, pawindow);
		x_delsloot(pawindow->titlebar, titlebar_handle, pawindow);
		XSelectInput(display, pawindow->frame, NoEventMask);
		XSelectInput(display, pawindow->titlebar, NoEventMask);
	}
	x_delsloot(pawindow->canvas, canvas_handle, pawindow);
	x_delsloot(pawindow->window, canvas_handle, pawindow);

	/* No more events */
	XSelectInput(display, pawindow->canvas, NoEventMask);

	/* Reparent the button to a safe place before destroying ;)), 
	 * but make sure that no other pawindow already did the reparenting */
	if (!XQueryTree(display,button_getwindow(class.button.close),
			&root,&parent,&tree,&nchildren)) BUG();

	if (tree) XFree(tree);

	if (pawindow->titlebar==parent) {
		button_unmap(class.button.close);
		button_unmap(class.button.iconz);
		button_unmap(class.button.maximize);
		button_reparent(class.button.close, ROOT, 0, 0);
		button_reparent(class.button.iconz, ROOT, 0, 0);
		button_reparent(class.button.maximize, ROOT, 0, 0);
	}

	/* Give back window client to the root */
	XReparentWindow(display, pawindow->window, ROOT, pawindow->x, pawindow->y);
	XDestroyWindow(display, pawindow->frame);
	if (pawindow->decoration && class.shadow_status)
		XDestroyWindow(display, pawindow->shadow);
	pabar_dellabel(pawindow->icon);
	palabel_destroy(pawindow->icon);

	/* Delete pawindow from chains */
	chain_home(class.pawindows);
	while ( (pw=chain_next(class.pawindows))!=NULL ) {
		if (pw==pawindow) {
			chain_del(class.pawindows); 
			break; 
		}
	}

	chain_home(class.fpawindows);
	while( (pw=chain_next(class.fpawindows))!=NULL ) {
		if (pw==pawindow) {
			chain_del(class.fpawindows);
		}
	}

    
	XFreePixmap(display, pawindow->wmicon->pixmap);
	XFreePixmap(display, pawindow->wmicon->pixmap);

	/* Helps debugging... */
	memset(pawindow, '\0', PAWindowBytes);
	free(pawindow);

	pawindow=NULL;

	/* Make sure nobody changed focuspawindow before. Damn global variables.*/
	if (!chain_size(class.fpawindows)) 
		focuspawindow=NULL;

	/* Update client windows list */
	set_client_list();

	/* Maybe there are no more windows mapped, so we reassure pabar is on top */
	pabar_map();


}


/* This maps and raises the pawindow and its decorations */
void pawindow_map (PAWindow pawindow)
{
	CARD32 data[2];   /* Fucking X portable types */
			  
	ASSERT(pawindow);

	SAY("Someone wants to map %#lx", pawindow->window);
	pawindow->active=True;

	/* A window is mapped and root window should release its grabs */
	keyboard_ungrabkeys(ROOT);
	
	/* A new pawindow is in withdrawn state, ICCCM-friendly */
	if ( (pawindow->state==PA_WITHDRAWN) && (!pawindow->got_skiptaskbar) &&
		 (!pawindow->utility) ) 
		pabar_addlabel(pawindow->icon);
	 
	/* Reparent and map buttons _before_ the pawindow, to make buttons draw
	 * faster when the window is mapped */   
	if (pawindow->decoration) {

		button_reparent(class.button.close, pawindow->titlebar,
						(int) TITLEBAR_W-BUTTONSZ, 0);
		button_action(class.button.close, kill_client, pawindow);
		button_reparent(class.button.maximize, pawindow->titlebar,
				(int) TITLEBAR_W-(BUTTONSZ*2), 0);
		button_action(class.button.maximize, maximize_client, pawindow);
		button_reparent(class.button.iconz, pawindow->titlebar,
						(int) TITLEBAR_W-(BUTTONSZ*3), 0);
		button_action(class.button.iconz, iconify_client, pawindow);

		button_map(class.button.close);
		button_map(class.button.iconz);
		button_map(class.button.maximize);
	} else {  /* No decoration, so buttons need to be put at a safe place ;) */
		if (!pawindow->utility) { /* Leave the buttons alone in utility windows */
			button_unmap(class.button.close);
			button_unmap(class.button.iconz);
			button_unmap(class.button.maximize);
			button_reparent(class.button.close, ROOT, 0, 0);
			button_reparent(class.button.iconz, ROOT, 0, 0);
			button_reparent(class.button.maximize, ROOT, 0, 0);
		}
	}

	/* Check fullscreen state */
	if (pawindow->fullscreen) 
		set_fullscreen(pawindow, true);
	else
		set_fullscreen(pawindow,false);

	/* Map the shadow window first, and resize it after mapping the client */
	if (pawindow->decoration && class.shadow_status) {
		XResizeWindow(display, pawindow->shadow, 1, 1);
		XMapRaised(display, pawindow->shadow);
	}

	if (!pawindow->shaded) {   
		XMapWindow(display, pawindow->window);
		XMapSubwindows(display, pawindow->frame);
	} else {
		XMapWindow(display, pawindow->frame);
		XMapWindow(display, pawindow->titlebar);
	}
	/* Map window two times, for slow drawing apps. First is the real map, 
	 * second time only is raised */
	XMapRaised(display, pawindow->frame);

	/* Keep above modal windows */
	if (pawindow->modalchild) {
		if (pawindow->modalchild->state!=PA_WITHDRAWN) {
			XMapSubwindows(display, pawindow->modalchild->frame);
			XMapRaised(display, pawindow->modalchild->frame);
		}
	}

	if (pawindow->decoration && class.shadow_status) 
		XResizeWindow(display, pawindow->shadow, SHADOW_W, SHADOW_H);

	draw_border(pawindow);
	draw_titlebar(pawindow);

	/* Raise the pabar if needed */
	if (!pawindow->fullscreen) 
		pabar_map();

	pawindow->state=PA_MAPPED;
	change_focus(pawindow);

	/* ICCCM, set the WM_STATE property */
	data[0]=NormalState;
	data[1]=None;
	x_set_property(pawindow->window, atoms.wm_state, XA_CARDINAL, (unsigned char *)&data, 2);
			
}




/* Unmaps the pawindow, no tricks. By now... }:)  */
void pawindow_unmap (PAWindow pawindow)
{
	CARD32 data[2];    /* Fucking X portable types */
	PAWindow pw;

	ASSERT(pawindow);

	SAY("Someone wants to unmap %#lx", pawindow->window);
	pawindow->state=PA_ICONIFIED;

	if (pawindow->decoration && class.shadow_status)
		XUnmapWindow(display, pawindow->shadow);
    
	XUnmapSubwindows(display, pawindow->frame);
	XUnmapWindow(display, pawindow->frame);
    
	if (pawindow->active==False) return;

	if (pawindow->icon) palabel_unsetfocus(pawindow->icon);

	/* Unmap modal windows attached to this one */
	if (pawindow->modalchild) {
		/* Do not modify withdrawn windows state */
		if (pawindow->modalchild->state==PA_MAPPED)
			pawindow->modalchild->state=PA_ICONIFIED;
		if (pawindow->modalchild->decoration && class.shadow_status)
			XUnmapWindow(display, pawindow->modalchild->shadow);
		
		XUnmapSubwindows(display, pawindow->modalchild->frame);
		XUnmapWindow(display, pawindow->modalchild->frame);
	}
		
    
	/* Get a new pawindow to be raised. FIXME: Get this outta here */
	chain_home(class.fpawindows);
	while( (pw=chain_next(class.fpawindows))!=NULL ) {
		if ( (pw!=pawindow) && (pw->state!=PA_WITHDRAWN) && (pw->state!=PA_ICONIFIED)) break; /* Gotcha */
	}

	if (pw!=NULL)
		pawindow_map(pw);
	else {
		/* No mapped pawindows */
		XSetInputFocus(display, ROOT, RevertToPointerRoot, CurrentTime);
		focuspawindow=NULL;
		keyboard_grabkeys(ROOT);
		pabar_map();
	}

	/* ICCCM, set the WM_STATE property */
	data[0]=IconicState;
	data[1]=None;
	x_set_property(pawindow->window, atoms.wm_state, XA_CARDINAL, (unsigned char *)&data, 2);

}



/****************************************************************************/






/*º Geometry management procedures XDDD *************************************/

/* Resize a PAWindow. Notice that 'width' and 'height' are client relative*/
void pawindow_resize (PAWindow pawindow, unsigned int w, unsigned int h) 
{
	ASSERT(pawindow);

	/* Honor client size limits, but don't give a shit for shaded windows */
	w=(w > pawindow->max_w) ? pawindow->max_w : w;
	h=(h > pawindow->max_h) ? pawindow->max_h : h;
	/* TODO: min_ limits are changing after window creation. See gucharmap */
	//w=(w < pawindow->min_w) ? pawindow->min_w : w;
	//h=(h < pawindow->min_h) ? pawindow->min_h : h;

	/* Change titlebar size */
	if (pawindow->decoration)
		XResizeWindow(display, pawindow->titlebar, w, TITLEBAR_H);

	/* Displace buttons with the bar resize, but only if the resized
	 * pawindow has the focus */
	if (pawindow==focuspawindow) {
		button_move(class.button.close, (int) w-BUTTONSZ, 0);
		button_move(class.button.maximize, (int) w-(BUTTONSZ*2), 0);
		button_move(class.button.iconz, (int) w-(BUTTONSZ*3), 0);
	}

	/* Change client's size */
	SAY("Resizing window to %d %d", w, h);
	XResizeWindow(display, pawindow->window, w, h);
	XResizeWindow(display, pawindow->canvas, w, h);

	if (pawindow->decoration) {
	   w+=BORDERW*2;
	   h+=BORDERW*2+TITLEBAR_H;
	}

	pawindow->w=w;
	pawindow->h=h;

	notify_client(pawindow);

	/* And the other windows too */
	if (pawindow->decoration && class.shadow_status)
		XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);


	XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

}




/* Move a PAWindow.*/
/* NOTE: 'x' and 'y' are supposed to be client relative... */
void pawindow_move (PAWindow pawindow, int x, int y) 
{
	int abs_x,abs_y,abs_h,abs_w;

	ASSERT(pawindow);

	if (pawindow->decoration) {
		abs_x=x-BORDERW; 
		abs_y=y-BORDERW-TITLEBAR_H;
		abs_w=pawindow->w-BORDERW*2;
		abs_h=pawindow->h-BORDERW*2-TITLEBAR_H;
	} else { 
		abs_x=x;
		abs_y=y;
		abs_w=pawindow->w; 
		abs_h=pawindow->h;
	}

	/* Sanitize values if the window is being moved out of sight */
	if (!pawindow->fullscreen) {
		if (abs_x < -abs_w ) abs_x=-abs_w+1;    
		if (abs_y < 1) abs_y=0;
	}

	pawindow->x=abs_x; pawindow->y=abs_y;    

	notify_client(pawindow);

	XMoveWindow(display, pawindow->frame, abs_x, abs_y);

	if (pawindow->decoration && class.shadow_status)
		XMoveWindow(display, pawindow->shadow, SHADOW_X, SHADOW_Y);

}


/* Shade (Make only visible the titlebar) a pawindow*/
void pawindow_shade (PAWindow pawindow) 
{
	ASSERT(pawindow);

	pawindow->shaded=True;
	pawindow->oldsh=pawindow->h;
	pawindow->h=TITLEBAR_H+(2*BORDERW);

	/* Unmap client windows*/
	XUnmapWindow(display, pawindow->window);
	XUnmapWindow(display, pawindow->canvas);

	/* Resize shadow */
	if (class.shadow_status)
		XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);

	/* And finally resize the frame */
	XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

	change_focus(pawindow);

}


/* Unshade a pawindow*/
void pawindow_unshade (PAWindow pawindow) 
{
	ASSERT(pawindow);

	pawindow->shaded=False;
	pawindow->h=pawindow->oldsh;

	/* Resize and map again the subwindows */   
	XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

	XMapWindow(display, pawindow->canvas);
	XMapWindow(display, pawindow->window);

	if (class.shadow_status)
		XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);

	/* Inform the client about the new position */
	notify_client(pawindow);

	change_focus(pawindow);
    
}

/****************************************************************************/




/*º Accessors****************************************************************/

/* This function returns a reference to the PAWindow icon */
PALabel pawindow_icon (PAWindow pawindow)
{
	ASSERT(pawindow);
	ASSERT(pawindow->icon);

	return pawindow->icon;
}

int pawindow_getwindowx (PAWindow pawindow) 
{
	ASSERT(pawindow);

	return pawindow->x+BORDERW;
}

int pawindow_getwindowy (PAWindow pawindow)
{
	ASSERT(pawindow);

	return pawindow->y+BORDERW+TITLEBAR_H;
}
/****************************************************************************/
